from .base import Item
from .member import Followers, Image
from .paging import CursorPaging, OffsetPaging


class Artist(Item):
    """Artist base."""

    external_urls: dict
    name: str


class SimpleArtist(Artist):
    """Simplified artist object."""


class FullArtist(Artist):
    """Complete artist object."""

    followers: Followers
    genres: list[str]
    images: list[Image]
    popularity: int


class FullArtistCursorPaging(CursorPaging):
    """Paging of full artists."""

    items: list[FullArtist]
    total: int


class FullArtistOffsetPaging(OffsetPaging):
    """Paging of full artists."""

    items: list[FullArtist]



from __future__ import annotations

from tekore._sender import Sender

from .expiring import AccessToken, Credentials, Token
from .scope import Scope


class RefreshingToken(AccessToken):
    """
    Automatically refreshing access token.

    Returned from utility functions and :class:`RefreshingCredentials`.
    It shouldn't have to be instantiated outside of the functions,
    unless you are sure that you want to.

    Uses an instance of :class:`Credentials` to automatically request
    a new access token when the old one is about to expire.
    This occurs when the :attr:`access_token` property is read.

    Both :attr:`expires_in` and :attr:`expires_at` are always ``None``,
    and :attr:`is_expiring` is always ``False``.

    Parameters
    ----------
    token
        access token object
    credentials
        credentials manager for token refreshing

    Attributes
    ----------
    credentials
        credentials manager for token refreshing
    """

    def __init__(self, token: Token, credentials: Credentials) -> None:
        self._token = token
        self.credentials = credentials

    def __repr__(self) -> str:
        options = [
            f"access_token={self.access_token!r}",
            f"refresh_token={self.refresh_token!r}",
            f"scope={self.scope!r}",
        ]
        return type(self).__name__ + "(" + ", ".join(options) + ")"

    @property
    def access_token(self) -> str:
        """Bearer token value."""
        if self._token.is_expiring:
            self._token = self.credentials.refresh(self._token)

        return self._token.access_token

    @property
    def refresh_token(self) -> str | None:
        """
        Refresh token for generating new access tokens.

        ``None`` if the token is an application token.
        """
        return self._token.refresh_token

    @property
    def token_type(self) -> str:
        """How the token may be used, always 'Bearer'."""
        return self._token.token_type

    @property
    def scope(self) -> Scope:
        """
        Privileges granted to the token.

        Empty :class:`Scope` if the token is an application token
        or a user token without any scopes.
        """
        return self._token.scope

    @property
    def expires_in(self) -> None:
        """Seconds until token expiration, always ``None``."""
        return None

    @property
    def expires_at(self) -> None:
        """When the token expires, always ``None``."""
        return None

    @property
    def is_expiring(self) -> bool:
        """Determine whether token is about to expire, always ``False``."""
        return False

    @property
    def uses_pkce(self) -> bool:
        """Proof key for code exchange used in authorisation."""
        return self._token.uses_pkce


class RefreshingCredentials:
    """
    Synchronous client for self-refreshing tokens.

    Delegates to an underlying :class:`Credentials` manager
    and parses tokens it returns into :class:`RefreshingToken`.

    Parameters
    ----------
    client_id
        client id
    client_secret
        client secret, not required for PKCE user authorisation
    redirect_uri
        whitelisted redirect URI, required for user authorisation
    sender
        synchronous request sender

    Attributes
    ----------
    credentials
        underlying credentials manager for token refreshing
    """

    def __init__(
        self,
        client_id: str,
        client_secret: str | None = None,
        redirect_uri: str | None = None,
        sender: Sender | None = None,
    ) -> None:
        self.credentials = Credentials(
            client_id, client_secret, redirect_uri, sender, asynchronous=False
        )

    def __repr__(self) -> str:
        options = [
            f"client_id={self.credentials.client_id!r}",
            f"client_secret={self.credentials.client_secret!r}",
            f"redirect_uri={self.credentials.redirect_uri!r}",
            f"sender={self.credentials.sender!r}",
        ]
        return type(self).__name__ + "(" + ", ".join(options) + ")"

    def request_client_token(self) -> RefreshingToken:
        """
        Request a refreshing client token.

        Returns
        -------
        RefreshingToken
            automatically refreshing client token
        """
        token = self.credentials.request_client_token()
        return RefreshingToken(token, self.credentials)

    def user_authorisation_url(
        self, scope=None, state: str | None = None, show_dialog: bool = False
    ) -> str:
        """
        Construct an authorisation URL.

        Step 1/2 in authorisation code flow.
        User should be redirected to the resulting URL for authorisation.
        Step 2/2: :meth:`request_user_token`.

        Parameters
        ----------
        scope
            token privileges, accepts a :class:`Scope`, a single :class:`scope`,
            a list of :class:`scopes <scope>` and strings for :class:`Scope`,
            or a space-separated list of scopes as a string
        state
            additional state
        show_dialog
            force login dialog even if previously authorised

        Returns
        -------
        str
            login URL
        """
        return self.credentials.user_authorisation_url(scope, state, show_dialog)

    def request_user_token(self, code: str) -> RefreshingToken:
        """
        Request a new refreshing user token.

        Step 2/2 in authorisation code flow.
        Code is provided as a URL parameter in the redirect URI
        after login in step 1: :meth:`user_authorisation_url`.

        Parameters
        ----------
        code
            code from redirect parameters

        Returns
        -------
        RefreshingToken
            automatically refreshing user token
        """
        token = self.credentials.request_user_token(code)
        return RefreshingToken(token, self.credentials)

    def refresh_user_token(self, refresh_token: str) -> RefreshingToken:
        """
        Request an automatically refreshing user token with a refresh token.

        Parameters
        ----------
        refresh_token
            refresh token

        Returns
        -------
        RefreshingToken
            automatically refreshing user token
        """
        token = self.credentials.refresh_user_token(refresh_token)
        return RefreshingToken(token, self.credentials)

    def pkce_user_authorisation(
        self, scope=None, state: str | None = None, verifier_bytes: int = 32
    ) -> tuple[str, str]:
        """
        Construct authorisation URL and verifier.

        Step 1/2 in authorisation code flow with proof key for code exchange.
        The user should be redirected to the resulting URL for authorisation.
        The verifier is passed to :meth:`request_pkce_token` in step 2.

        Parameters
        ----------
        scope
            token privileges, accepts a :class:`Scope`, a single :class:`scope`,
            a list of :class:`scopes <scope>` and strings for :class:`Scope`,
            or a space-separated list of scopes as a string
        state
            additional state
        verifier_bytes
            number of bytes to generate PKCE verifier with, ``32 <= bytes <= 96``.
            The specified range of bytes generates the appropriate number of
            characters (43 - 128) after base-64 encoding, as required in RFC 7636.

        Returns
        -------
        tuple[str, str]
            authorisation URL and PKCE code verifier
        """
        return self.credentials.pkce_user_authorisation(scope, state, verifier_bytes)

    def request_pkce_token(self, code: str, verifier: str) -> RefreshingToken:
        """
        Request a new PKCE user token.

        Step 2/2 in authorisation code flow with proof key for code exchange.
        Code is provided as a URL parameter in the redirect URI
        after login in step 1: :meth:`pkce_user_authorisation`.

        Parameters
        ----------
        code
            code from redirect parameters
        verifier
            PKCE code verifier generated for authorisation URL

        Returns
        -------
        RefreshingToken
            user access token
        """
        token = self.credentials.request_pkce_token(code, verifier)
        return RefreshingToken(token, self.credentials)

    def refresh_pkce_token(self, refresh_token: str) -> RefreshingToken:
        """
        Request a refreshed PKCE user token.

        Parameters
        ----------
        refresh_token
            refresh token

        Returns
        -------
        RefreshingToken
            refreshed user access token
        """
        token = self.credentials.refresh_pkce_token(refresh_token)
        return RefreshingToken(token, self.credentials)



from tekore._auth import scope
from tekore._client.base import SpotifyBase
from tekore._client.decor import scopes, send_and_process
from tekore._client.process import single
from tekore.model import PrivateUser, PublicUser


class SpotifyUser(SpotifyBase):
    """User API endpoints."""

    @scopes()
    @send_and_process(single(PublicUser))
    def user(self, user_id: str) -> PublicUser:
        """
        Get a user's profile.

        Parameters
        ----------
        user_id
            user ID
        """
        return self._get("users/" + user_id.replace("#", "%23"))

    @scopes(optional=[scope.user_read_private, scope.user_read_email])
    @send_and_process(single(PrivateUser))
    def current_user(self) -> PrivateUser:
        """
        Get current user's profile.

        The user-read-private scope allows the user's country and
        product subscription level to be returned.
        """
        return self._get("me/")



from __future__ import annotations

from tekore._model.episode import SimpleEpisodePaging
from tekore._model.show import Show


class FullShow(Show):
    """
    Complete show object.

    :attr:`total_episodes` is undocumented by Spotify,
    so it might be missing or removed in a future version.
    """

    episodes: SimpleEpisodePaging | None = None



from __future__ import annotations

import webbrowser
from secrets import token_urlsafe
from urllib.parse import parse_qs, urlparse

from .expiring import Credentials, Token
from .refreshing import RefreshingCredentials, RefreshingToken


def gen_state(n_bytes: int = 32) -> str:
    """
    Generate state to use in user authorisation.

    The generated state is random and URL-safe.
    It is generated using :func:`secrets.token_urlsafe`.
    """
    return token_urlsafe(n_bytes)


def _parse_url_param(url: str, param: str) -> str:
    query = urlparse(url).query
    code = parse_qs(query).get(param, None)

    if code is None:
        msg = f"Passed URL contains no parameter `{param}`!"
        raise KeyError(msg)
    if len(code) > 1:
        msg = f"Passed URL contains multiple values for `{param}`!"
        raise KeyError(msg)

    return code[0]


def parse_code_from_url(url: str) -> str:
    """
    Parse an URL for parameter 'code'.

    Returns
    -------
    str
        value of 'code'

    Raises
    ------
    KeyError
        if 'code' is not available or has multiple values
    """
    return _parse_url_param(url, "code")


def parse_state_from_url(url: str) -> str:
    """
    Parse an URL for parameter 'state'.

    Returns
    -------
    str
        value of 'state'

    Raises
    ------
    KeyError
        if 'state' is not available or has multiple values
    """
    return _parse_url_param(url, "state")


class UserAuth:
    """
    Implement user authorisation flow.

    Implements all steps and security checks for user authorisation.
    The responsibility of the caller is to redirect a user to the given URL
    and provide the resulting redirect URI or its parameters.
    Can be used with an asynchronous credentials client.

    Parameters
    ----------
    cred
        credentials client
    scope
        token privileges, accepts a :class:`Scope`, a single :class:`scope`,
        a list of :class:`scopes <scope>` and strings for :class:`Scope`,
        or a space-separated list of scopes as a string
    pkce
        use proof key for code exchange

    Attributes
    ----------
    url: str
        address to redirect a user to for authorisation
    state: str
        generated additional state
    verifier: str
        PKCE code verifier, :class:`None` if PKCE is not used

    Examples
    --------
    .. code:: python

        auth = tk.UserAuth(cred, scope)

        # Redirect user to auth.url and parse parameters
        code, state = ...
        token = auth.request_token(code, state)

        # Or leave parsing to UserAuth
        redirected = ...
        token = auth.request_token(url=redirected)

        # With an asynchronous client
        token = await auth.request_token(url=redirected)
    """

    def __init__(
        self, cred: Credentials | RefreshingCredentials, scope=None, pkce: bool = False
    ) -> None:
        self._cred = cred
        self.state = gen_state()
        self.verifier = None
        if pkce:
            self.url, self.verifier = self._cred.pkce_user_authorisation(
                scope, self.state
            )
        else:
            self.url = self._cred.user_authorisation_url(
                scope, self.state, show_dialog=True
            )

    def __repr__(self) -> str:
        options = [
            f"cred={self._cred!r}",
            f"url={self.url!r}",
            f"state={self.state!r}",
            f"verifier={self.verifier}",
        ]
        return type(self).__name__ + "(" + ", ".join(options) + ")"

    def request_token(
        self, code: str | None = None, state: str | None = None, url: str | None = None
    ) -> Token | RefreshingToken:
        """
        Verify state consistency and request token.

        Parameters
        ----------
        code
            code from redirect parameters, required if url was not specified
        state
            state from redirect parameters, required if url was not specified
        url
            if specified, code and state are parsed from this URL instead

        Returns
        -------
        Token | RefreshingToken
            access token

        Raises
        ------
        AssertionError
            if state is inconsistent
        """
        if url is not None:
            code = parse_code_from_url(url)
            state = parse_state_from_url(url)

        if self.state != state:
            msg = f"Inconsistent state! Expected `{self.state}`, got `{state}`."
            raise AssertionError(msg)

        if self.verifier is not None:
            return self._cred.request_pkce_token(code, self.verifier)
        return self._cred.request_user_token(code)


def request_client_token(client_id: str, client_secret: str) -> RefreshingToken:
    """
    Request for client credentials.

    Parameters
    ----------
    client_id
        client ID
    client_secret
        client secret

    Returns
    -------
    RefreshingToken
        automatically refreshing client token
    """
    cred = RefreshingCredentials(client_id, client_secret)
    return cred.request_client_token()


def prompt_for_user_token(
    client_id: str,
    client_secret: str,
    redirect_uri: str,
    scope=None,
    open_browser: bool = True,
) -> RefreshingToken:
    """
    Prompt for manual authorisation.

    Open a web browser for the user to log in with Spotify.
    Prompt to paste the URL after logging in to complete authorisation.

    Parameters
    ----------
    client_id
        client ID
    client_secret
        client secret
    redirect_uri
        whitelisted redirect URI
    scope
        token privileges, accepts a :class:`Scope`, a single :class:`scope`,
        a list of :class:`scopes <scope>` and strings for :class:`Scope`,
        or a space-separated list of scopes as a string
    open_browser
        open a web browser with auth url, or just print it

    Returns
    -------
    RefreshingToken
        automatically refreshing user token

    Raises
    ------
    AssertionError
        if state is inconsistent
    """
    cred = RefreshingCredentials(client_id, client_secret, redirect_uri)
    auth = UserAuth(cred, scope=scope)

    if open_browser:
        print("Opening browser for Spotify login...")  # noqa: T201
        webbrowser.open(auth.url)
    else:
        print("Open this URL in your browser: " + auth.url)  # noqa: T201
    redirected = input("Paste the final redirected URL from the browser: ").strip()
    return auth.request_token(url=redirected)


def refresh_user_token(
    client_id: str, client_secret: str, refresh_token: str
) -> RefreshingToken:
    """
    Request a refreshed user token.

    Parameters
    ----------
    client_id
        client ID
    client_secret
        client secret
    refresh_token
        refresh token

    Returns
    -------
    RefreshingToken
        automatically refreshing user token
    """
    cred = RefreshingCredentials(client_id, client_secret)
    return cred.refresh_user_token(refresh_token)


def prompt_for_pkce_token(
    client_id: str, redirect_uri: str, scope=None, open_browser: bool = True
) -> RefreshingToken:
    """
    Prompt for manual authorisation with PKCE.

    Open a web browser for the user to log in with Spotify.
    Prompt to paste the URL after logging in to complete authorisation.

    Parameters
    ----------
    client_id
        client ID
    redirect_uri
        whitelisted redirect URI
    scope
        token privileges, accepts a :class:`Scope`, a single :class:`scope`,
        a list of :class:`scopes <scope>` and strings for :class:`Scope`,
        or a space-separated list of scopes as a string
    open_browser
        open a web browser with auth url, or just print it

    Returns
    -------
    RefreshingToken
        automatically refreshing PKCE user token

    Raises
    ------
    AssertionError
        if state is inconsistent
    """
    cred = RefreshingCredentials(client_id, redirect_uri=redirect_uri)
    auth = UserAuth(cred, scope=scope, pkce=True)

    if open_browser:
        print("Opening browser for Spotify login...")  # noqa: T201
        webbrowser.open(auth.url)
    else:
        print("Open this URL in your browser: " + auth.url)  # noqa: T201
    redirected = input("Paste the final redirected URL from the browser: ").strip()
    return auth.request_token(url=redirected)


def refresh_pkce_token(client_id: str, refresh_token: str) -> RefreshingToken:
    """
    Request a refreshed PKCE user token.

    Parameters
    ----------
    client_id
        client ID
    refresh_token
        refresh token

    Returns
    -------
    RefreshingToken
        automatically refreshing user token
    """
    cred = RefreshingCredentials(client_id)
    return cred.refresh_pkce_token(refresh_token)



from .serialise import Model, StrEnum


class ContextType(StrEnum):
    """Type of player context."""

    album = "album"
    artist = "artist"
    collection = "collection"
    playlist = "playlist"
    show = "show"


class Context(Model):
    """Context of a played track or episode."""

    type: ContextType
    href: str
    external_urls: dict
    uri: str



from enum import Enum


class PlayerErrorReason(Enum):
    """Reasons for errors in player actions."""

    NO_PREV_TRACK = (
        "The command requires a previous track, but there is none in the context."
    )
    NO_NEXT_TRACK = (
        "The command requires a next track, but there is none in the context."
    )
    NO_SPECIFIC_TRACK = "The requested track does not exist."
    ALREADY_PAUSED = "The command requires playback to not be paused."
    NOT_PAUSED = "The command requires playback to be paused."
    NOT_PLAYING_LOCALLY = "The command requires playback on the local device."
    NOT_PLAYING_TRACK = "The command requires that a track is currently playing."
    NOT_PLAYING_CONTEXT = "The command requires that a context is currently playing."
    ENDLESS_CONTEXT = "The shuffle command cannot be applied on an endless context."
    CONTEXT_DISALLOW = "The command could not be performed on the context."
    ALREADY_PLAYING = (
        "The track should not be restarted if the same track "
        "and context is already playing, and there is a resume point."
    )
    RATE_LIMITED = (
        "The user is rate limited due to too frequent track play,"
        "also known as cat-on-the-keyboard spamming."
    )
    REMOTE_CONTROL_DISALLOW = "The context cannot be remote-controlled."
    DEVICE_NOT_CONTROLLABLE = "Not possible to remote control the device."
    VOLUME_CONTROL_DISALLOW = "Not possible to remote control the device's volume."
    NO_ACTIVE_DEVICE = "Requires an active device and the user has none."
    PREMIUM_REQUIRED = "The request is prohibited for non-premium users."
    UNKNOWN = "Certain actions are restricted because of unknown reasons."



from __future__ import annotations

from collections.abc import Callable, Iterable
from functools import wraps

from tekore._auth import scope
from tekore._client.base import SpotifyBase
from tekore._client.chunked import _get_arg
from tekore._client.decor import maximise_limit, scopes, send_and_process
from tekore._client.process import model_list, nothing, single
from tekore.model import FullPlaylist, Image, PlaylistTrackPaging, SimplePlaylistPaging


def process_if_not_specified(post_func: Callable, *arguments) -> Callable:
    """
    Decorate a function to process only if any of arguments is falsy.

    Parameters
    ----------
    post_func
        function to call with response JSON content
    arguments
        arguments to check, tuples of (name, position in argument list)
    """

    def decorator(function: Callable) -> Callable:
        async def async_wrapper(self: SpotifyBase, *args, **kwargs):
            json = await function(self, *args, **kwargs)
            return post_func(json)

        @wraps(function)
        def wrapper(self: SpotifyBase, *args, **kwargs):
            falsies = [
                bool(_get_arg(arg_pos - 1, arg_name, args, kwargs))
                for arg_name, arg_pos in arguments
            ]
            if any(falsies):
                return function(self, *args, **kwargs)

            if self.is_async:
                return async_wrapper(self, *args, **kwargs)

            json = function(self, *args, **kwargs)
            return post_func(json)

        return wrapper

    return decorator


def parse_additional_types(as_tracks: bool | Iterable[str]) -> str | None:
    """Determine `additional_types` argument content."""
    types = {"track", "episode"}
    if as_tracks is True:
        types = set()
    elif as_tracks is False:
        pass
    else:
        types = types.difference(as_tracks)

    return ",".join(types) if types else None


class SpotifyPlaylistView(SpotifyBase):
    """Playlist API endpoints for viewing playlists."""

    @scopes(optional=[scope.playlist_read_private, scope.playlist_read_collaborative])
    @send_and_process(single(SimplePlaylistPaging))
    @maximise_limit(50)
    def followed_playlists(
        self, limit: int = 20, offset: int = 0
    ) -> SimplePlaylistPaging:
        """
        Get a list of the playlists owned or followed by the current user.

        Parameters
        ----------
        limit
            the number of items to return (1..50)
        offset
            the index of the first item to return
        """
        return self._get("me/playlists", limit=limit, offset=offset)

    @scopes(optional=[scope.playlist_read_private, scope.playlist_read_collaborative])
    @send_and_process(single(SimplePlaylistPaging))
    @maximise_limit(50)
    def playlists(
        self, user_id: str, limit: int = 20, offset: int = 0
    ) -> SimplePlaylistPaging:
        """
        Get a list of the playlists owned or followed by a user.

        Collaborative playlists are only returned for the current user.

        Parameters
        ----------
        user_id
            user ID
        limit
            the number of items to return (1..50)
        offset
            the index of the first item to return
        """
        return self._get(f"users/{user_id}/playlists", limit=limit, offset=offset)

    @scopes()
    @process_if_not_specified(single(FullPlaylist), ("fields", 2), ("as_tracks", 4))
    @send_and_process(nothing)
    def playlist(
        self,
        playlist_id: str,
        fields: str | None = None,
        market: str | None = None,
        as_tracks: bool | Iterable[str] = False,
    ) -> FullPlaylist | dict:
        """
        Get playlist of a user.

        .. note:: Returns a dictionary if ``fields`` or ``as_tracks`` is specified.

        Parameters
        ----------
        playlist_id
            playlist ID
        fields
            which fields to return, see the Web API documentation for details
        market
            an ISO 3166-1 alpha-2 country code or 'from_token'
            when using a user token to authenticate.
            For episodes in the playlist, if a user token is used,
            the country associated with it overrides this parameter.
            If an application token is used and no market is specified,
            episodes are considered unavailable and returned as None.
        as_tracks
            return types of items with track-like fields.
            If :class:`True`, return all other types as tracks.
            If an iterable is passed, types contained are returned as tracks.
            Currently the only extra type is ``episode``.

        Returns
        -------
        FullPlaylist | dict
            playlist object, or raw dictionary if ``fields``
            or ``as_tracks`` was specified
        """
        additional_types = parse_additional_types(as_tracks)
        return self._get(
            "playlists/" + playlist_id,
            fields=fields,
            market=market,
            additional_types=additional_types,
        )

    @scopes()
    @send_and_process(model_list(Image))
    def playlist_cover_image(self, playlist_id: str) -> list[Image]:
        """
        Get cover image of a playlist.

        .. note:: Returns a list of images.

        Parameters
        ----------
        playlist_id
            playlist ID
        """
        return self._get(f"playlists/{playlist_id}/images")

    @scopes()
    @process_if_not_specified(
        single(PlaylistTrackPaging), ("fields", 2), ("as_tracks", 4)
    )
    @send_and_process(nothing)
    def playlist_items(
        self,
        playlist_id: str,
        fields: str | None = None,
        market: str | None = None,
        as_tracks: bool | Iterable[str] = False,
        limit: int = 100,
        offset: int = 0,
    ) -> PlaylistTrackPaging | dict:
        """
        Full details of items on a playlist.

        .. note:: Returns a dictionary if ``fields`` or ``as_tracks`` is specified.

        Parameters
        ----------
        playlist_id
            playlist ID
        fields
            which fields to return, see the Web API documentation for details
        market
            an ISO 3166-1 alpha-2 country code or 'from_token'
            when using a user token to authenticate.
            For episodes in the playlist, if a user token is used,
            the country associated with it overrides this parameter.
            If an application token is used and no market is specified,
            episodes are considered unavailable and returned as None.
        as_tracks
            return types of items with track-like fields.
            If :class:`True`, return all other types as tracks.
            If an iterable is passed, types contained are returned as tracks.
            Currently the only extra type is ``episode``.
        limit
            the number of items to return (1..100)
        offset
            the index of the first item to return

        Returns
        -------
        PlaylistTrackPaging | dict
            paging object containing playlist items, or raw dictionary
            if ``fields`` or ``as_tracks`` was specified
        """
        return self._get(
            f"playlists/{playlist_id}/tracks",
            limit=limit,
            offset=offset,
            fields=fields,
            market=market,
            additional_types=parse_additional_types(as_tracks),
        )



from __future__ import annotations

import re

from tekore.model import StrEnum


class ConversionError(Exception):
    """Error in conversion."""


class IdentifierType(StrEnum):
    """Valid types of Spotify IDs."""

    artist = "artist"
    album = "album"
    episode = "episode"
    playlist = "playlist"
    show = "show"
    track = "track"
    user = "user"


def check_type(type_: str | IdentifierType) -> None:
    """
    Validate type of an ID.

    Raises
    ------
    ConversionError
        When type is invalid.
    """
    if str(type_) not in IdentifierType.__members__:
        msg = f"Invalid type {type_!r}!"
        raise ConversionError(msg)


# Match beginning, all base62 characters and end of string
all_base62 = re.compile("^[0-9a-zA-Z]*$")


def check_id(id_: str) -> None:
    """
    Validate resource ID to be base 62.

    Note that user IDs can have special characters, so they cannot be validated.

    Raises
    ------
    ConversionError
        When ID is invalid.
    """
    if id_ == "" or all_base62.search(id_) is None:
        msg = f"Invalid id: {id_!r}!"
        raise ConversionError(msg)


def to_uri(type_: str | IdentifierType, id_: str) -> str:
    """
    Convert an ID to an URI of the appropriate type.

    Parameters
    ----------
    type_
        valid :class:`IdentifierType`
    id_
        resource identifier

    Returns
    -------
    str
        converted URI

    Raises
    ------
    ConversionError
        On invalid type or ID.
    """
    check_type(type_)
    if type_ != IdentifierType.user:
        check_id(id_)
    return f"spotify:{type_}:{id_}"


def to_url(type_: str | IdentifierType, id_: str) -> str:
    """
    Convert an ID to an URL of the appropriate type.

    Parameters
    ----------
    type_
        valid :class:`IdentifierType`
    id_
        resource identifier

    Returns
    -------
    str
        converted URL

    Raises
    ------
    ConversionError
        On invalid type or ID.
    """
    check_type(type_)
    if type_ != IdentifierType.user:
        check_id(id_)
    else:
        id_ = id_.replace("#", "%23")
    return f"https://open.spotify.com/{type_}/{id_}"


def from_uri(uri: str) -> tuple[str, str]:
    """
    Parse type and ID from an URI.

    Parameters
    ----------
    uri
        URI to parse

    Returns
    -------
    tuple[str, str]
        type and ID parsed from the URI

    Raises
    ------
    ConversionError
        On invalid format, prefix, type or ID.
    """
    try:
        type_, id_ = _parse_uri(uri)
    except ValueError as e:
        msg = f'Invalid URI: expected format "spotify:{{type}}:{{id}}", got {uri!r}!'
        raise ConversionError(msg) from e

    check_type(type_)
    if type_ != IdentifierType.user:
        check_id(id_)

    return type_, id_


def _parse_uri(uri: str) -> tuple[str, str]:
    spotify, type_, id_ = uri.split(":")
    if spotify != "spotify":
        raise ValueError
    return type_, id_


_url_prefixes = (
    "open.spotify.com",
    "http://open.spotify.com",
    "https://open.spotify.com",
)


def from_url(url: str) -> tuple[str, str]:
    """
    Parse type and ID from an URL.

    Any parameters in the URL will be ignored.

    Parameters
    ----------
    url
        URL to parse

    Returns
    -------
    tuple[str, str]
        type and ID parsed from the URL

    Raises
    ------
    ConversionError
        On invalid format, prefix, type or ID.
    """
    try:
        type_, id_ = _parse_url(url)
    except ValueError as e:
        valid_url = "[http[s]://]open.spotify.com/{type}/{id}"
        msg = f"Invalid URL: expected format {valid_url!r}, got {url!r}!"
        raise ConversionError(msg) from e

    id_ = id_.split("?")[0]
    check_type(type_)
    if type_ != IdentifierType.user:
        check_id(id_)

    return type_, id_


def _parse_url(url: str) -> tuple[str, str]:
    *prefixes, type_, id_ = url.split("/")
    prefix = "/".join(prefixes)
    if prefix not in _url_prefixes:
        raise ValueError
    return type_, id_



from __future__ import annotations

from tekore._auth import scope
from tekore._client.base import SpotifyBase
from tekore._client.decor import scopes, send_and_process
from tekore._client.process import nothing
from tekore._convert import to_uri
from tekore.model import RepeatState


def offset_to_dict(offset: int | str | None) -> dict | None:
    """
    Parse playback start offset to an appropriate payload member.

    If offset is an integer, it is an index to a track position.
    If it is a string, it is a URI of a specific track.
    """
    if isinstance(offset, int):
        return {"position": offset}
    if isinstance(offset, str):
        return {"uri": to_uri("track", offset)}
    return None


class SpotifyPlayerModify(SpotifyBase):
    """Player API endpoints that modify state."""

    @scopes([scope.user_modify_playback_state])
    @send_and_process(nothing)
    def playback_transfer(self, device_id: str, force_play: bool = False) -> None:
        """
        Transfer playback to another device.

        Parameters
        ----------
        device_id
            device to transfer playback to
        force_play
            true: play after transfer, false: keep current state
        """
        payload = {"device_ids": [device_id], "play": force_play}
        return self._put("me/player", payload=payload)

    @scopes([scope.user_modify_playback_state])
    @send_and_process(nothing)
    def playback_resume(self, device_id: str | None = None) -> None:
        """
        Resume user's playback.

        Parameters
        ----------
        device_id
            device to start playback on
        """
        return self._put("me/player/play", device_id=device_id)

    @scopes([scope.user_modify_playback_state])
    @send_and_process(nothing)
    def playback_start_tracks(
        self,
        track_ids: list[str],
        offset: int | str | None = None,
        position_ms: int | None = None,
        device_id: str | None = None,
    ) -> None:
        """
        Start playback of one or more tracks.

        Parameters
        ----------
        track_ids
            track IDs to start playing
        offset
            offset into tracks by index or track ID
        position_ms
            initial position of first played track
        device_id
            device to start playback on
        """
        payload = {
            "uris": [to_uri("track", t) for t in track_ids],
            "offset": offset_to_dict(offset),
            "position_ms": position_ms,
        }
        payload = {k: v for k, v in payload.items() if v is not None}
        return self._put("me/player/play", payload=payload, device_id=device_id)

    @scopes([scope.user_modify_playback_state])
    @send_and_process(nothing)
    def playback_start_context(
        self,
        context_uri: str,
        offset: int | str | None = None,
        position_ms: int | None = None,
        device_id: str | None = None,
    ) -> None:
        """
        Start playback of a context: an album, artist or playlist.

        Parameters
        ----------
        context_uri
            context to start playing
        offset
            offset into context by index or track ID,
            only available when context is an album or playlist
        position_ms
            initial position of first played track
        device_id
            device to start playback on
        """
        payload = {
            "context_uri": context_uri,
            "offset": offset_to_dict(offset),
            "position_ms": position_ms,
        }
        payload = {k: v for k, v in payload.items() if v is not None}
        return self._put("me/player/play", payload=payload, device_id=device_id)

    @scopes([scope.user_modify_playback_state])
    @send_and_process(nothing)
    def playback_queue_add(self, uri: str, device_id: str | None = None) -> None:
        """
        Add a track or an episode to a user's queue.

        Parameters
        ----------
        uri
            resource to add, track or episode
        device_id
            devide to extend the queue on
        """
        return self._post("me/player/queue", uri=uri, device_id=device_id)

    @scopes([scope.user_modify_playback_state])
    @send_and_process(nothing)
    def playback_pause(self, device_id: str | None = None) -> None:
        """
        Pause a user's playback.

        Parameters
        ----------
        device_id
            device to pause playback on
        """
        return self._put("me/player/pause", device_id=device_id)

    @scopes([scope.user_modify_playback_state])
    @send_and_process(nothing)
    def playback_next(self, device_id: str | None = None) -> None:
        """
        Skip user's playback to next track.

        Parameters
        ----------
        device_id
            device to skip track on
        """
        return self._post("me/player/next", device_id=device_id)

    @scopes([scope.user_modify_playback_state])
    @send_and_process(nothing)
    def playback_previous(self, device_id: str | None = None) -> None:
        """
        Skip user's playback to previous track.

        Parameters
        ----------
        device_id
            device to skip track on
        """
        return self._post("me/player/previous", device_id=device_id)

    @scopes([scope.user_modify_playback_state])
    @send_and_process(nothing)
    def playback_seek(self, position_ms: int, device_id: str | None = None) -> None:
        """
        Seek to position in current playing track.

        Parameters
        ----------
        position_ms
            position on track
        device_id
            device to seek on
        """
        return self._put("me/player/seek", position_ms=position_ms, device_id=device_id)

    @scopes([scope.user_modify_playback_state])
    @send_and_process(nothing)
    def playback_repeat(
        self, state: str | RepeatState, device_id: str | None = None
    ) -> None:
        """
        Set repeat mode for playback.

        Parameters
        ----------
        state
            `track`, `context`, or `off`
        device_id
            device to set repeat on
        """
        return self._put("me/player/repeat", state=str(state), device_id=device_id)

    @scopes([scope.user_modify_playback_state])
    @send_and_process(nothing)
    def playback_shuffle(self, state: bool, device_id: str | None = None) -> None:
        """
        Toggle shuffle for user's playback.

        Parameters
        ----------
        state
            shuffle state
        device_id
            device to toggle shuffle on
        """
        state_arg = "true" if state else "false"
        return self._put("me/player/shuffle", state=state_arg, device_id=device_id)

    @scopes([scope.user_modify_playback_state])
    @send_and_process(nothing)
    def playback_volume(
        self, volume_percent: int, device_id: str | None = None
    ) -> None:
        """
        Set volume for user's playback.

        Parameters
        ----------
        volume_percent
            volume to set (0..100)
        device_id
            device to set volume on
        """
        return self._put(
            "me/player/volume", volume_percent=volume_percent, device_id=device_id
        )



from .album import SpotifyAlbum
from .artist import SpotifyArtist
from .audiobook import SpotifyAudiobook
from .browse import SpotifyBrowse
from .chapter import SpotifyChapter
from .episode import SpotifyEpisode
from .follow import SpotifyFollow
from .library import SpotifyLibrary
from .markets import SpotifyMarkets
from .personalisation import SpotifyPersonalisation
from .player import SpotifyPlayer
from .playlist import SpotifyPlaylist
from .search import SpotifySearch
from .show import SpotifyShow
from .track import SpotifyTrack
from .user import SpotifyUser



from __future__ import annotations

from tekore._client.base import SpotifyBase
from tekore._client.chunked import chunked, join_lists
from tekore._client.decor import maximise_limit, scopes, send_and_process
from tekore._client.process import model_list, single
from tekore.model import FullAudiobook, SimpleChapterPaging


class SpotifyAudiobook(SpotifyBase):
    """Audiobook API endpoints."""

    @scopes()
    @send_and_process(single(FullAudiobook))
    def audiobook(self, audiobook_id: str, market: str | None = None) -> FullAudiobook:
        """
        Get information for an audiobook.

        Parameters
        ----------
        audiobook_id
            audiobook ID
        market
            an ISO 3166-1 alpha-2 country code.
            If a user token is used to authenticate, the country associated
            with it overrides this parameter.
            If an application token is used and no market is specified,
            the show is considered unavailable.
        """
        return self._get("audiobooks/" + audiobook_id, market=market)

    @scopes()
    @chunked("audiobook_ids", 1, 50, join_lists)
    @send_and_process(model_list(FullAudiobook, "audiobooks"))
    def audiobooks(
        self, audiobook_ids: list[str], market: str | None = None
    ) -> list[FullAudiobook]:
        """
        Get information for multiple audiobooks.

        Parameters
        ----------
        audiobook_ids
            the audiobook IDs, max 50 without chunking
        market
            an ISO 3166-1 alpha-2 country code.
            If a user token is used to authenticate, the country associated
            with it overrides this parameter.
            If an application token is used and no market is specified,
            the show is considered unavailable.
        """
        return self._get("audiobook", ids=",".join(audiobook_ids), market=market)

    @scopes()
    @send_and_process(single(SimpleChapterPaging))
    @maximise_limit(50)
    def audiobook_chapters(
        self,
        audiobook_id: str,
        market: str | None = None,
        limit: int = 20,
        offset: int = 0,
    ) -> SimpleChapterPaging:
        """
        Get chapters of an audiobook.

        Parameters
        ----------
        audiobook_id
            audiobook ID
        market
            an ISO 3166-1 alpha-2 country code.
            If a user token is used to authenticate, the country associated
            with it overrides this parameter.
            If an application token is used and no market is specified,
            the show is considered unavailable.
        limit
            the number of items to return (1..50)
        offset
            the index of the first item to return
        """
        return self._get(
            f"audiobooks/{audiobook_id}/chapters",
            market=market,
            limit=limit,
            offset=offset,
        )



from httpx import codes

from .base import Request, Response


class HTTPError(Exception):
    """
    Base error for all web status errors.

    Attributes
    ----------
    request
        request that led to the error
    response
        response from the web server
    """

    def __init__(self, message: str, request: Request, response: Response) -> None:
        super().__init__(message)
        self.request = request
        self.response = response


class ClientError(HTTPError):
    """4xx - Base client error."""


class ServerError(HTTPError):
    """5xx - Base server error."""


class BadRequest(ClientError):
    """
    400 - Bad request.

    The request could not be understood by the server due to malformed syntax.
    """


class Unauthorised(ClientError):
    """
    401 - Unauthorised.

    The request requires user authentication or,
    if the request included authorization credentials,
    authorization has been refused for those credentials.

    The scopes associated with the call are attached to this class.
    """

    scope: str
    required_scope: str
    optional_scope: str


class Forbidden(ClientError):
    """
    403 - Forbidden.

    The server understood the request, but is refusing to fulfill it.
    """


class NotFound(ClientError):
    """
    404 - Not found.

    The requested resource could not be found.
    This error can be due to a temporary or permanent condition.
    """


class TooManyRequests(ClientError):
    """
    429 - Too many requests.

    Rate limiting has been applied.
    """


class InternalServerError(ServerError):
    """
    500 - Internal server error.

    You should never receive this error because the clever coders at Spotify
    catch them all... But if you are unlucky enough to get one,
    please report it to Spotify through their GitHub (spotify/web-api).
    """


class BadGateway(ClientError):
    """
    502 - Bad gateway.

    The server was acting as a gateway or proxy and received
    an invalid response from the upstream server.
    """


class ServiceUnavailable(ClientError):
    """
    503 - Service unavailable.

    The server is currently unable to handle the request due to a temporary
    condition which will be alleviated after some delay.
    You can choose to resend the request again.
    """


errors = {
    400: BadRequest,
    401: Unauthorised,
    403: Forbidden,
    404: NotFound,
    429: TooManyRequests,
    500: InternalServerError,
    502: BadGateway,
    503: ServiceUnavailable,
}


def get_error(code: int) -> type[HTTPError]:
    """Get error based on status code or default error."""
    cls = errors.get(code)
    if cls is None:
        cls = ClientError if codes.is_client_error(code) else ServerError
    return cls



from __future__ import annotations

from datetime import datetime
from typing import Literal

from .base import Item
from .episode import FullEpisode
from .local import LocalTrack
from .member import Followers, Image
from .paging import OffsetPaging
from .serialise import Model
from .track import FullTrack, Tracks
from .user import PublicUser


class FullPlaylistTrack(FullTrack):
    """
    Track on a playlist.

    Provides :attr:`episode` and :attr:`track` booleans
    to easily determine the type of playlist item.
    """

    episode: Literal[False]
    track: Literal[True]
    is_local: Literal[False]


class FullPlaylistEpisode(FullEpisode):
    """
    Episode on a playlist.

    Provides :attr:`episode` and :attr:`track` booleans
    to easily determine the type of playlist item.
    :attr:`available_markets` is undocumented.
    """

    available_markets: list[str] | None = None
    episode: Literal[True]
    track: Literal[False]


class LocalPlaylistTrack(LocalTrack):
    """
    Local track on a playlist.

    Provides :attr:`episode` and :attr:`track` booleans
    to easily determine the type of playlist item.
    """

    episode: Literal[False] = False
    track: Literal[True] = True


class PlaylistTrack(Model):
    """Track or episode on a playlist."""

    added_at: datetime
    added_by: PublicUser
    is_local: bool
    track: FullPlaylistTrack | FullPlaylistEpisode | LocalPlaylistTrack | None

    primary_color: str | None
    video_thumbnail: dict | None


class PlaylistTrackPaging(OffsetPaging):
    """Paging of playlist tracks."""

    items: list[PlaylistTrack]


class Playlist(Item):
    """
    Playlist base.

    :attr:`owner` can be ``None`` on featured playlists.
    """

    collaborative: bool
    description: str | None
    external_urls: dict
    images: list[Image] | None
    name: str
    owner: PublicUser
    public: bool | None
    snapshot_id: str

    primary_color: str | None


class SimplePlaylist(Playlist):
    """Simplified playlist object."""

    tracks: Tracks


class FullPlaylist(Playlist):
    """Complete playlist object."""

    followers: Followers
    tracks: PlaylistTrackPaging


class SimplePlaylistPaging(OffsetPaging):
    """Paging of simplified playlists."""

    items: list[SimplePlaylist | None]



from collections.abc import Callable

from httpx import codes

from tekore._sender import Request, Response
from tekore._sender.error import get_error

from .token import Token


def handle_errors(request: Request, response: Response) -> None:
    """Examine response and raise errors accordingly."""
    if not codes.is_error(response.status_code):
        return

    if codes.is_client_error(response.status_code):
        error_str = f"{response.status_code} {response.content['error']}"
        description = response.content.get("error_description", None)
        if description is not None:
            error_str += ": " + description
    else:
        error_str = "Unexpected error!"

    error_cls = get_error(response.status_code)
    raise error_cls(error_str, request=request, response=response)


def parse_token(uses_pkce: bool) -> Callable:
    """Wrap token parsing conditional to PKCE usage."""

    def func(request: Request, response: Response) -> Token:
        """Parse token object from response."""
        handle_errors(request, response)
        return Token(response.content, uses_pkce)

    return func


def parse_refreshed_token(uses_pkce: bool) -> Callable:
    """Wrap token parsing conditional to PKCE usage."""

    def func(request: Request, response: Response, refresh_token: str) -> Token:
        """Replace new refresh token with old value if empty."""
        handle_errors(request, response)
        refreshed = Token(response.content, uses_pkce)

        if refreshed.refresh_token is None:
            refreshed._refresh_token = refresh_token

        return refreshed

    return func



from __future__ import annotations

from tekore._model.chapter import SimpleChapterPaging

from .base import Audiobook


class FullAudiobook(Audiobook):
    """Complete audiobook object."""

    chapters: SimpleChapterPaging
    is_playable: bool | None = None



from __future__ import annotations

from pydantic import Field

from tekore._model.base import Item
from tekore._model.member import Copyright, Image


class Show(Item):
    """
    Show base.

    :attr:`publisher` being an object is undocumented.
    """

    available_markets: list[str] = Field(default_factory=list)
    copyrights: list[Copyright]
    description: str
    explicit: bool
    external_urls: dict
    html_description: str | None = None
    images: list[Image]
    is_externally_hosted: bool | None
    languages: list[str]
    media_type: str
    name: str
    publisher: str | dict
    total_episodes: int | None = None



from __future__ import annotations

from collections.abc import Callable, Coroutine
from functools import wraps
from warnings import warn

from .base import Request, Response
from .concrete import AsyncSender, Sender, SyncSender
from .error import Unauthorised
from .extending import ExtendingSender


class SenderConflictWarning(RuntimeWarning):
    """Sender arguments to a client are in conflict."""


class Client(ExtendingSender):
    """
    Base class for clients.

    Parameters
    ----------
    sender
        request sender - If not specified, a :class:`SyncSender` is used
    asynchronous
        synchronicity requirement - If specified, overrides passed sender
        if they are in conflict and instantiates a sender of the requested type
    """

    def __init__(self, sender: Sender | None, asynchronous: bool | None = None) -> None:
        super().__init__(sender)

        if self.sender.is_async and asynchronous is False:
            self.sender = SyncSender()
        elif not self.sender.is_async and asynchronous is True:
            self.sender = AsyncSender()

        if sender is not None and self.sender.is_async != sender.is_async:
            msg = (
                f"{type(sender)} with is_async={sender.is_async} passed"
                f" but asynchronous={asynchronous}!"
                f"\nA new {type(self.sender).__name__} was instantiated."
            )
            warn(msg, SenderConflictWarning, stacklevel=3)

    def send(self, request: Request) -> Response | Coroutine[None, None, Response]:
        """Send request with underlying sender."""
        return self.sender.send(request)


def send_and_process(post_func: Callable) -> Callable:
    """
    Decorate a Client function to send a request and process its content.

    The first parameter of a decorated function must be the instance (self)
    of a :class:`Sender` (has :meth:`send` and :attr:`is_async`).
    The decorated function must return a tuple with two items:
    a :class:`Request` and a tuple with arguments to unpack to ``post_func``.
    The result of ``post_func`` is returned to the caller.

    Parameters
    ----------
    post_func
        function to call with the request and response
        and possible additional arguments
    """

    def decorator(function: Callable[..., Request]) -> Callable:
        def try_post_func(*args, **kwargs):
            try:
                return post_func(*args, **kwargs)
            except Unauthorised as e:
                e.scope = wrapper.scope
                e.required_scope = wrapper.required_scope
                e.optional_scope = wrapper.optional_scope
                raise

        async def async_send(self: Sender, request: Request, params: tuple):
            response = await self.send(request)
            return try_post_func(request, response, *params)

        @wraps(function)
        def wrapper(self: Sender, *args, **kwargs):
            request, params = function(self, *args, **kwargs)

            if self.is_async:
                return async_send(self, request, params)

            response = self.send(request)
            return try_post_func(request, response, *params)

        return wrapper

    return decorator



from __future__ import annotations

import asyncio
import time
from collections import deque
from collections.abc import Coroutine
from urllib.parse import urlencode

from httpx import codes

from .base import Request, Response
from .concrete import Sender, SyncSender


class ExtendingSender(Sender):
    """
    Base class for senders that extend other senders.

    Parameters
    ----------
    sender
        request sender, :class:`SyncSender` if not specified
    """

    def __init__(self, sender: Sender | None) -> None:
        self.sender = sender or SyncSender()

    @property
    def is_async(self) -> bool:
        """Sender asynchronicity, delegated to the underlying sender."""
        return self.sender.is_async

    def close(self) -> None | Coroutine[None, None, None]:
        """
        Close the underlying sender.

        To close synchronous senders, call :meth:`close`.
        To close asynchronous senders, await :meth:`close`.
        """
        return self.sender.close()


class RetryingSender(ExtendingSender):
    """
    Retry requests if unsuccessful.

    On server errors the set amount of retries are used to resend requests.
    On :class:`TooManyRequests` the `Retry-After` header is checked and used
    to wait before requesting again.

    .. note::

        Even when the number of retries is set to zero,
        retries based on rate limiting are still performed.

    Parameters
    ----------
    retries
        maximum number of retries on server errors before giving up
    sender
        request sender, :class:`SyncSender` if not specified

    Examples
    --------
    Use for only rate limiting by leaving the retry count to zero.

    .. code:: python

        tk.RetryingSender()

    Pass the maximum number of retries to retry failed requests.

    .. code:: python

        tk.RetryingSender(retries=3)
    """

    def __init__(self, retries: int = 0, sender: Sender | None = None) -> None:
        super().__init__(sender)
        self.retries = retries

    def __repr__(self) -> str:
        contains = f"(retries={self.retries}, sender={self.sender!r})"
        return type(self).__name__ + contains

    def send(self, request: Request) -> Response | Coroutine[None, None, Response]:
        """Delegate request to underlying sender and retry if failed."""
        if self.is_async:
            return self._async_send(request)

        retries = max(self.retries, 0)
        delay_seconds = 1

        while True:
            r = self.sender.send(request)

            if r.status_code == codes.TOO_MANY_REQUESTS:
                seconds = r.headers.get("Retry-After", 1)
                time.sleep(int(seconds) + 1)
            elif codes.is_server_error(r.status_code) and retries > 0:
                retries -= 1
                time.sleep(delay_seconds)
                delay_seconds *= 2
            else:
                break
        return r

    async def _async_send(self, request: Request) -> Response:
        retries = max(self.retries, 0)
        delay_seconds = 1

        while True:
            r = await self.sender.send(request)

            if r.status_code == codes.TOO_MANY_REQUESTS:
                seconds = r.headers.get("Retry-After", 1)
                await asyncio.sleep(int(seconds) + 1)
            elif codes.is_server_error(r.status_code) and retries > 0:
                retries -= 1
                await asyncio.sleep(delay_seconds)
                delay_seconds *= 2
            else:
                break
        return r


class CachingSender(ExtendingSender):
    """
    Cache successful GET requests.

    The Web API provides response headers for caching.
    Resources are cached based on Cache-Control, ETag and Vary headers.
    Thus :class:`CachingSender` can be used with user tokens too.
    Resources marked as private, errors and ``Vary: *`` are not cached.

    When using asynchronous senders, the cache is protected with
    :class:`asyncio.Lock` to prevent concurrent access.
    The lock is instantiated on the first asynchronous call,
    so using only one :func:`asyncio.run` (per sender) is advised.

    Note that if the cache has no maximum size it can grow without limit.
    Use :meth:`CachingSender.clear` to empty the cache.

    Parameters
    ----------
    max_size
        maximum cache size (amount of responses), if specified the least
        recently used response is discarded when the cache would overflow
    sender
        request sender, :class:`SyncSender` if not specified
    """

    def __init__(
        self, max_size: int | None = None, sender: Sender | None = None
    ) -> None:
        super().__init__(sender)
        self._max_size = max_size
        self._cache: dict[str, tuple] = {}
        self._deque: deque = deque(maxlen=self.max_size)
        self._lock: asyncio.Lock | None = None

    def __repr__(self) -> str:
        contains = f"(max_size={self._max_size}, sender={self.sender!r})"
        return type(self).__name__ + contains

    @property
    def max_size(self) -> int | None:
        """
        Maximum amount of requests stored in the cache.

        Returns
        -------
        int | None
            maximum cache size
        """
        return self._max_size

    def clear(self) -> None:
        """Clear sender cache."""
        self._cache = {}
        self._deque.clear()

    @staticmethod
    def _vary_key(request: Request, vary: list[str] | None) -> str | None:
        if vary is not None:
            return " ".join(request.headers[k] for k in vary)
        return None

    @staticmethod
    def _cc_fresh(item: dict) -> bool:
        return item["expires_at"] > time.time()

    @staticmethod
    def _has_etag(item: dict) -> bool:
        return item["etag"] is not None

    def _is_fresh(self, url: str, vary_key: str) -> bool:
        item = self._cache[url][1][vary_key]
        return self._cc_fresh(item) or self._has_etag(item)

    def _delete(self, url: str, vary_key: str) -> None:
        item = self._cache[url]
        del item[1][vary_key]
        if not item[1]:
            del item

    def _remove_stale_items(self) -> None:
        if len(self._deque) == self._deque.maxlen:
            deque_items = list(self._deque)
            self._deque.clear()

            for item in deque_items:
                fresh = self._is_fresh(*item)

                if fresh:
                    self._deque.append(item)
                else:
                    self._delete(*item)

    def _append_item(self, item: tuple[str, str | None]) -> None:
        # Remove LRU item
        if len(self._deque) == self._deque.maxlen:
            d_url, d_vary_key = self._deque.popleft()
            self._delete(d_url, d_vary_key)

        self._deque.append(item)

    def _maybe_save(self, request: Request, response: Response) -> None:
        cc = response.headers.get("Cache-Control", "private, max-age=0")

        if codes.is_error(response.status_code) or "private" in cc:
            return

        age = int(cc.split("max-age=")[1].split(",")[0])
        vary = response.headers.get("Vary", None)
        if vary is not None:
            if "*" in vary:
                return
            vary = vary.split(", ")

        # Construct cached response
        cache_item = self._cache.setdefault(response.url, (vary, {}))
        cached_response = {
            "response": response,
            "expires_at": time.time() + age - 1,
            "etag": response.headers.get("ETag", None),
        }
        vary_key = self._vary_key(request, vary)
        cache_item[1].update({vary_key: cached_response})

        # Manage cache size
        if self.max_size is None:
            return

        self._remove_stale_items()
        self._append_item((response.url, vary_key))

    def _update_usage(self, item: tuple[str, str | None]) -> None:
        if self.max_size is None:
            return

        self._deque.remove(item)
        self._deque.append(item)

    def _load(self, request: Request) -> tuple:
        params = ("&" + urlencode(request.params)) if request.params else ""
        url = request.url + params
        item = self._cache.get(url, None)

        if item is None:
            return None, None

        vary_key = self._vary_key(request, item[0])
        cached = item[1].get(vary_key, None)

        if cached is not None:
            response = cached["response"]
            deque_item = (url, vary_key)
            if self._cc_fresh(cached):
                self._update_usage(deque_item)
                return response, None
            if self._has_etag(cached):
                self._update_usage(deque_item)
                return response, cached["etag"]
            if self.max_size is not None:
                self._deque.remove(deque_item)

        return None, None

    def _handle_fresh(
        self, request: Request, fresh: Response, cached: Response
    ) -> Response:
        if fresh.status_code == codes.NOT_MODIFIED:
            return cached
        self._maybe_save(request, fresh)
        return fresh

    def send(self, request: Request) -> Response | Coroutine[None, None, Response]:
        """Maybe load request from cache, or delegate to underlying sender."""
        if self.is_async:
            return self._async_send(request)

        if request.method.lower() != "get":
            return self.sender.send(request)

        cached, etag = self._load(request)
        if cached is not None and etag is None:
            return cached
        if etag is not None:
            request.headers.update(ETag=etag)

        fresh = self.sender.send(request)
        return self._handle_fresh(request, fresh, cached)

    async def _async_send(self, request: Request) -> Response:
        if request.method.lower() != "get":
            return await self.sender.send(request)

        if self._lock is None:
            self._lock = asyncio.Lock()

        async with self._lock:
            cached, etag = self._load(request)

        if cached is not None and etag is None:
            return cached
        if etag is not None:
            request.headers.update(ETag=etag)

        fresh = await self.sender.send(request)
        async with self._lock:
            return self._handle_fresh(request, fresh, cached)



from .album import Album, AlbumGroup, AlbumType, SimpleAlbum, SimpleAlbumPaging
from .album.full import FullAlbum, SavedAlbum, SavedAlbumPaging
from .artist import (
    Artist,
    FullArtist,
    FullArtistCursorPaging,
    FullArtistOffsetPaging,
    SimpleArtist,
)
from .audio_analysis import AudioAnalysis, Section, Segment, TimeInterval
from .audio_features import AudioFeatures
from .audiobook import (
    Audiobook,
    Author,
    Narrator,
    SimpleAudiobook,
    SimpleAudiobookPaging,
)
from .audiobook.full import FullAudiobook
from .base import Identifiable, Item
from .category import Category, CategoryPaging
from .chapter import Chapter, SimpleChapter, SimpleChapterPaging
from .chapter.full import FullChapter
from .context import Context, ContextType
from .currently_playing import (
    Actions,
    CurrentlyPlaying,
    CurrentlyPlayingContext,
    CurrentlyPlayingType,
    Disallows,
    Queue,
    RepeatState,
)
from .device import Device, DeviceType
from .episode import (
    Episode,
    FullEpisode,
    SavedEpisode,
    SavedEpisodePaging,
    SimpleEpisode,
    SimpleEpisodePaging,
)
from .error import PlayerErrorReason
from .local import LocalAlbum, LocalArtist, LocalItem, LocalTrack
from .member import (
    Copyright,
    Followers,
    Image,
    ReleaseDatePrecision,
    Restrictions,
    ResumePoint,
)
from .paging import Cursor, CursorPaging, OffsetPaging, Paging
from .play_history import PlayHistory, PlayHistoryCursor, PlayHistoryPaging
from .playlist import (
    FullPlaylist,
    FullPlaylistEpisode,
    FullPlaylistTrack,
    LocalPlaylistTrack,
    Playlist,
    PlaylistTrack,
    PlaylistTrackPaging,
    SimplePlaylist,
    SimplePlaylistPaging,
)
from .recommendations import (
    RecommendationAttribute,
    Recommendations,
    RecommendationSeed,
)
from .serialise import Model, StrEnum, UnknownModelAttributeWarning
from .show import SavedShow, SavedShowPaging, Show, SimpleShow, SimpleShowPaging
from .show.full import FullShow
from .track import (
    FullTrack,
    FullTrackPaging,
    SavedTrack,
    SavedTrackPaging,
    SimpleTrack,
    SimpleTrackPaging,
    Track,
    TrackLink,
    Tracks,
)
from .user import ExplicitContent, PrivateUser, PublicUser, User



from __future__ import annotations

from tekore._client.base import SpotifyBase
from tekore._client.chunked import chunked, join_lists
from tekore._client.decor import scopes, send_and_process
from tekore._client.process import model_list, single
from tekore.model import FullEpisode


class SpotifyEpisode(SpotifyBase):
    """Episode API endpoints."""

    @scopes()
    @send_and_process(single(FullEpisode))
    def episode(self, episode_id: str, market: str | None = None) -> FullEpisode:
        """
        Get information for an episode.

        Parameters
        ----------
        episode_id
            episode ID
        market
            an ISO 3166-1 alpha-2 country code.
            If a user token is used to authenticate, the country associated
            with it overrides this parameter.
            If an application token is used and no market is specified,
            the episode is considered unavailable.
        """
        return self._get("episodes/" + episode_id, market=market)

    @scopes()
    @chunked("episode_ids", 1, 50, join_lists)
    @send_and_process(model_list(FullEpisode, "episodes"))
    def episodes(
        self, episode_ids: list[str], market: str | None = None
    ) -> list[FullEpisode]:
        """
        Get information for multiple episodes.

        Parameters
        ----------
        episode_ids
            the episode IDs, max 50 without chunking
        market
            an ISO 3166-1 alpha-2 country code.
            If a user token is used to authenticate, the country associated
            with it overrides this parameter.
            If an application token is used and no market is specified,
            the episode is considered unavailable.
        """
        return self._get("episodes", ids=",".join(episode_ids), market=market)



from __future__ import annotations

from tekore._client.base import SpotifyBase
from tekore._client.decor import maximise_limit, scopes, send_and_process
from tekore.model import (
    FullArtistOffsetPaging,
    FullTrackPaging,
    SimpleAlbumPaging,
    SimpleAudiobookPaging,
    SimpleEpisodePaging,
    SimplePlaylistPaging,
    SimpleShowPaging,
)

paging_type = {
    "artists": FullArtistOffsetPaging,
    "albums": SimpleAlbumPaging,
    "audiobooks": SimpleAudiobookPaging,
    "episodes": SimpleEpisodePaging,
    "playlists": SimplePlaylistPaging,
    "shows": SimpleShowPaging,
    "tracks": FullTrackPaging,
}


def search_result(json: dict):
    """Unpack search result dicts into respective paging type constructors."""
    return tuple(paging_type[key](**json[key]) for key in json)


class SpotifySearch(SpotifyBase):
    """Search API endpoints."""

    @scopes()
    @send_and_process(search_result)
    @maximise_limit(50)
    def search(
        self,
        query: str,
        types: tuple = ("track",),
        market: str | None = None,
        include_external: str | None = None,
        limit: int = 20,
        offset: int = 0,
    ) -> tuple:
        """
        Search for an item.

        Returns :class:`NotFound` if limit+offset would be above 1000.

        Parameters
        ----------
        query
            search query
        types
            resources to search for, tuple of strings containing
            artist, album, audiobook, track, playlist, show and/or episode
        market
            an ISO 3166-1 alpha-2 country code or 'from_token'
        limit
            the number of items to return (1..50)
        offset
            the index of the first item to return
        include_external
            if 'audio', response will include any externally hosted audio

        Returns
        -------
        tuple
            Paging objects containing the types of items searched for
            in the order that they were specified in 'types'.

            * artist: :class:`FullArtistOffsetPaging <model.FullArtistOffsetPaging>`
            * album: :class:`SimpleAlbumPaging <model.SimpleAlbumPaging>`
            * audiobook: :class:`SimpleAudiobookPaging <model.SimpleAudiobookPaging>`
            * episode: :class:`SimpleEpisodePaging <model.SimpleEpisodePaging>`
            * playlist: :class:`SimplePlaylistPaging <model.SimplePlaylistPaging>`
            * show: :class:`SimpleShowPaging <model.SimpleShowPaging>`
            * track: :class:`FullTrackPaging <model.FullTrackPaging>`

        Examples
        --------
        .. code:: python

            tracks, = spotify.search('monty python')
            artists, = spotify.search('sheeran', types=('artist',))
            albums, tracks = spotify.search('piano', types=('album', 'track'))
            spotify.search('gold album:boba artist:abba', types=('track',))
            spotify.search('bob year:1980-2020', types=('show',))

        .. note::
            You can narrow down search results by specifying field filters
            (e.g. year range, genre).
            See the `Search for an Item
            <https://developer.spotify.com/documentation/web-api/reference/>`_
            page of the official documentation for more information.

        """
        return self._get(
            "search",
            q=query,
            type=",".join(types),
            market=market,
            include_external=include_external,
            limit=limit,
            offset=offset,
        )



from __future__ import annotations

from tekore._model.artist import SimpleArtist
from tekore._model.base import Item
from tekore._model.member import Image, ReleaseDatePrecision
from tekore._model.serialise import StrEnum


class AlbumType(StrEnum):
    """Type of album."""

    album = "album"
    compilation = "compilation"
    single = "single"
    ep = "ep"


class Album(Item):
    """Album base."""

    album_type: AlbumType
    artists: list[SimpleArtist]
    external_urls: dict
    images: list[Image]
    name: str
    total_tracks: int
    release_date: str
    release_date_precision: ReleaseDatePrecision
    available_markets: list[str] | None = None
    is_playable: bool | None = None



from __future__ import annotations

from tekore._client.base import SpotifyBase
from tekore._client.chunked import chunked, join_lists
from tekore._client.decor import scopes, send_and_process
from tekore._client.process import model_list, single
from tekore.model import AudioAnalysis, AudioFeatures, FullTrack


class SpotifyTrack(SpotifyBase):
    """Track API endpoints."""

    @scopes()
    @send_and_process(single(FullTrack))
    def track(self, track_id: str, market: str | None = None) -> FullTrack:
        """
        Get information for a track.

        Parameters
        ----------
        track_id
            track ID
        market
            an ISO 3166-1 alpha-2 country code or 'from_token'
        """
        return self._get("tracks/" + track_id, market=market)

    @scopes()
    @chunked("track_ids", 1, 50, join_lists)
    @send_and_process(model_list(FullTrack, "tracks"))
    def tracks(
        self, track_ids: list[str], market: str | None = None
    ) -> list[FullTrack]:
        """
        Get information for multiple tracks.

        Parameters
        ----------
        track_ids
            the track IDs, max 50 without chunking
        market
            an ISO 3166-1 alpha-2 country code or 'from_token'
        """
        return self._get("tracks", ids=",".join(track_ids), market=market)

    @scopes()
    @send_and_process(single(AudioAnalysis))
    def track_audio_analysis(self, track_id: str) -> AudioAnalysis:
        """
        Get a detailed audio analysis for a track.

        .. warning::

            This endpoint is unavailable to new third-party applications (:issue:`331`)

        The analysis describes the track's structure and musical content,
        including rythm, pitch and timbre.
        """
        return self._get("audio-analysis/" + track_id)

    @scopes()
    @send_and_process(single(AudioFeatures))
    def track_audio_features(self, track_id: str) -> AudioFeatures:
        """
        Get audio feature information for a track.

        .. warning::

            This endpoint is unavailable to new third-party applications (:issue:`331`)
        """
        return self._get("audio-features/" + track_id)

    @scopes()
    @chunked("track_ids", 1, 100, join_lists)
    @send_and_process(model_list(AudioFeatures, "audio_features"))
    def tracks_audio_features(self, track_ids: list[str]) -> list[AudioFeatures]:
        """
        Get audio feature information for multiple tracks.

        .. warning::

            This endpoint is unavailable to new third-party applications (:issue:`331`)

        Feature information for a track may be ``None`` if not available.

        Parameters
        ----------
        track_ids
            track IDs, max 100 without chunking
        """
        return self._get("audio-features", ids=",".join(track_ids))



from __future__ import annotations

from httpx import AsyncClient, Client
from httpx import Response as HTTPXResponse

from .base import Request, Response, Sender


def try_parse_json(response: HTTPXResponse) -> dict | None:
    """Parse json content or return None if not successful."""
    try:
        return response.json()
    except ValueError:
        return None


class SyncSender(Sender):
    """
    Send requests synchronously.

    .. warning::

        The underlying client is *not* closed automatically.
        Use :code:`sender.client.close()` to close it,
        particularly if multiple senders are instantiated.

    Parameters
    ----------
    client
        :class:`httpx.Client` to use when sending requests
    """

    def __init__(self, client: Client | None = None) -> None:
        self.client = client or Client()

    def send(self, request: Request) -> Response:
        """Send request with :class:`httpx.Client`."""
        response = self.client.request(
            method=request.method,
            url=request.url,
            params=request.params,
            headers=request.headers,
            data=request.data,
            json=request.json,
            content=request.content,
        )
        return Response(
            url=str(response.url),
            headers=dict(response.headers),
            status_code=response.status_code,
            content=try_parse_json(response),
        )

    @property
    def is_async(self) -> bool:
        """Sender asynchronicity, always :class:`False`."""
        return False

    def close(self) -> None:
        """Close the underlying synchronous client."""
        return self.client.close()


class AsyncSender(Sender):
    """
    Send requests asynchronously.

    .. warning::

        The underlying client is **not** closed automatically.
        Use :code:`await sender.client.aclose()` to close it,
        particularly if multiple senders are instantiated.

    Parameters
    ----------
    client
        :class:`httpx.AsyncClient` to use when sending requests
    """

    def __init__(self, client: AsyncClient | None = None) -> None:
        self.client = client or AsyncClient()

    async def send(self, request: Request) -> Response:
        """Send request with :class:`httpx.AsyncClient`."""
        response = await self.client.request(
            method=request.method,
            url=request.url,
            params=request.params,
            headers=request.headers,
            data=request.data,
            json=request.json,
            content=request.content,
        )
        return Response(
            url=str(response.url),
            headers=dict(response.headers),
            status_code=response.status_code,
            content=try_parse_json(response),
        )

    @property
    def is_async(self) -> bool:
        """Sender asynchronicity, always :class:`True`."""
        return True

    async def close(self) -> None:
        """Close the underlying asynchronous client."""
        return await self.client.aclose()



from __future__ import annotations

from tekore._auth import scope
from tekore._client.base import SpotifyBase
from tekore._client.chunked import chunked, join_lists, return_none
from tekore._client.decor import maximise_limit, scopes, send_and_process
from tekore._client.process import nothing, single
from tekore.model import (
    SavedAlbumPaging,
    SavedEpisodePaging,
    SavedShowPaging,
    SavedTrackPaging,
)


class SpotifyLibrary(SpotifyBase):
    """Library API endpoints."""

    @scopes([scope.user_library_read])
    @send_and_process(single(SavedAlbumPaging))
    @maximise_limit(50)
    def saved_albums(
        self, market: str | None = None, limit: int = 20, offset: int = 0
    ) -> SavedAlbumPaging:
        """
        Get the albums saved in the current user's library.

        Parameters
        ----------
        market
            an ISO 3166-1 alpha-2 country code or 'from_token'
        limit
            the number of items to return (1..50)
        offset
            the index of the first item to return
        """
        return self._get("me/albums", market=market, limit=limit, offset=offset)

    @scopes([scope.user_library_read])
    @chunked("album_ids", 1, 50, join_lists)
    @send_and_process(nothing)
    def saved_albums_contains(self, album_ids: list[str]) -> list[bool]:
        """
        Check if user has saved albums.

        Parameters
        ----------
        album_ids
            list of album IDs, max 50 without chunking

        Returns
        -------
        list[bool]
            save statuses in the same order the album IDs were given
        """
        return self._get("me/albums/contains", ids=",".join(album_ids))

    @scopes([scope.user_library_modify])
    @chunked("album_ids", 1, 50, return_none)
    @send_and_process(nothing)
    def saved_albums_add(self, album_ids: list[str]) -> None:
        """
        Save albums for current user.

        Parameters
        ----------
        album_ids
            list of album IDs, max 50 without chunking
        """
        return self._put("me/albums", ids=",".join(album_ids))

    @scopes([scope.user_library_modify])
    @chunked("album_ids", 1, 50, return_none)
    @send_and_process(nothing)
    def saved_albums_delete(self, album_ids: list[str]) -> None:
        """
        Remove albums for current user.

        Parameters
        ----------
        album_ids
            list of album IDs, max 50 without chunking
        """
        return self._delete("me/albums", ids=",".join(album_ids))

    @scopes([scope.user_library_read])
    @send_and_process(single(SavedEpisodePaging))
    @maximise_limit(50)
    def saved_episodes(
        self, market: str | None = None, limit: int = 20, offset: int = 0
    ) -> SavedEpisodePaging:
        """
        Get the episodes saved in the current user's library.

        Parameters
        ----------
        market
            an ISO 3166-1 alpha-2 country code or 'from_token'
        limit
            the number of items to return (1..50)
        offset
            the index of the first item to return
        """
        return self._get("me/episodes", market=market, limit=limit, offset=offset)

    @scopes([scope.user_library_read])
    @chunked("episode_ids", 1, 50, join_lists)
    @send_and_process(nothing)
    def saved_episodes_contains(self, episode_ids: list[str]) -> list[bool]:
        """
        Check if user has saved episodes.

        Parameters
        ----------
        episode_ids
            list of episode IDs, max 50 without chunking

        Returns
        -------
        list[bool]
            save statuses in the same order the episode IDs were given
        """
        return self._get("me/episodes/contains", ids=",".join(episode_ids))

    @scopes([scope.user_library_modify])
    @chunked("episode_ids", 1, 50, return_none)
    @send_and_process(nothing)
    def saved_episodes_add(self, episode_ids: list[str]) -> None:
        """
        Save episodes for current user.

        Parameters
        ----------
        episode_ids
            list of episode IDs, max 50 without chunking
        """
        return self._put("me/episodes", ids=",".join(episode_ids))

    @scopes([scope.user_library_modify])
    @chunked("episode_ids", 1, 50, return_none)
    @send_and_process(nothing)
    def saved_episodes_delete(self, episode_ids: list[str]) -> None:
        """
        Remove episodes for current user.

        Parameters
        ----------
        episode_ids
            list of episode IDs, max 50 without chunking
        """
        return self._delete("me/episodes", ids=",".join(episode_ids))

    @scopes([scope.user_library_read])
    @send_and_process(single(SavedTrackPaging))
    @maximise_limit(50)
    def saved_tracks(
        self, market: str | None = None, limit: int = 20, offset: int = 0
    ) -> SavedTrackPaging:
        """
        Get the songs saved in the current user's library.

        Parameters
        ----------
        market
            an ISO 3166-1 alpha-2 country code or 'from_token'
        limit
            the number of items to return (1..50)
        offset
            the index of the first item to return
        """
        return self._get("me/tracks", market=market, limit=limit, offset=offset)

    @scopes([scope.user_library_read])
    @chunked("track_ids", 1, 50, join_lists)
    @send_and_process(nothing)
    def saved_tracks_contains(self, track_ids: list[str]) -> list[bool]:
        """
        Check if user has saved tracks.

        Parameters
        ----------
        track_ids
            list of track IDs, max 50 without chunking

        Returns
        -------
        list[bool]
            save statuses in the same order the track IDs were given
        """
        return self._get("me/tracks/contains", ids=",".join(track_ids))

    @scopes([scope.user_library_modify])
    @chunked("track_ids", 1, 50, return_none)
    @send_and_process(nothing)
    def saved_tracks_add(self, track_ids: list[str]) -> None:
        """
        Save tracks for current user.

        Parameters
        ----------
        track_ids
            list of track IDs, max 50 without chunking
        """
        return self._put("me/tracks", ids=",".join(track_ids))

    @scopes([scope.user_library_modify])
    @chunked("track_ids", 1, 50, return_none)
    @send_and_process(nothing)
    def saved_tracks_delete(self, track_ids: list[str]) -> None:
        """
        Remove tracks for current user.

        Parameters
        ----------
        track_ids
            list of track IDs, max 50 without chunking
        """
        return self._delete("me/tracks", ids=",".join(track_ids))

    @scopes([scope.user_library_read])
    @send_and_process(single(SavedShowPaging))
    @maximise_limit(50)
    def saved_shows(
        self, market: str | None = None, limit: int = 20, offset: int = 0
    ) -> SavedShowPaging:
        """
        Get the shows saved in the current user's library.

        Parameters
        ----------
        market
            an ISO 3166-1 alpha-2 country code or 'from_token'
        limit
            the number of items to return (1..50)
        offset
            the index of the first item to return
        """
        return self._get("me/shows", market=market, limit=limit, offset=offset)

    @scopes([scope.user_library_read])
    @chunked("show_ids", 1, 50, join_lists)
    @send_and_process(nothing)
    def saved_shows_contains(self, show_ids: list[str]) -> list[bool]:
        """
        Check if user has saved shows.

        Parameters
        ----------
        show_ids
            list of show IDs, max 50 without chunking

        Returns
        -------
        list[bool]
            save statuses in the same order the show IDs were given
        """
        return self._get("me/shows/contains", ids=",".join(show_ids))

    @scopes([scope.user_library_modify])
    @chunked("show_ids", 1, 50, return_none)
    @send_and_process(nothing)
    def saved_shows_add(self, show_ids: list[str]) -> None:
        """
        Save shows for current user.

        Parameters
        ----------
        show_ids
            list of show IDs, max 50 without chunking
        """
        return self._put("me/shows", ids=",".join(show_ids))

    @scopes([scope.user_library_modify])
    @chunked("show_ids", 1, 50, return_none)
    @send_and_process(nothing)
    def saved_shows_delete(
        self, show_ids: list[str], market: str | None = None
    ) -> None:
        """
        Remove shows for current user.

        Parameters
        ----------
        show_ids
            list of show IDs, max 50 without chunking
        market
            an ISO 3166-1 alpha-2 country code, only remove shows that are
            available in the specified market, overrided by token's country
        """
        return self._delete("me/shows", ids=",".join(show_ids), market=market)



from __future__ import annotations

from collections.abc import Generator

from tekore._sender import BadRequest
from tekore.model import Model, OffsetPaging, Paging

from .base import SpotifyBase
from .decor import send_and_process


def parse_paging_result(result: dict) -> dict:
    """Parse through the varying paging layouts."""
    # If only one top-level key, the paging object is one level deeper
    if len(result) == 1:
        key = next(iter(result.keys()))
        result = result[key]

    return result


class SpotifyPaging(SpotifyBase):
    """Paging navigation endpoints."""

    @send_and_process(parse_paging_result)
    def _get_paging_result(self, address: str):
        return self._get(address)

    def next(self, page: Paging) -> Paging | None:
        """
        Retrieve the next result set of a paging object.

        Parameters
        ----------
        page
            paging object

        Returns
        -------
        Paging
            next result set
        """
        if self.is_async:
            return self._async_next(page)

        if page.next is None:
            return None

        try:
            next_set = self._get_paging_result(page.next)
            return type(page)(**next_set)
        except BadRequest:
            return None

    async def _async_next(self, page: Paging) -> Paging | None:
        if page.next is None:
            return None

        try:
            next_set = await self._get_paging_result(page.next)
            return type(page)(**next_set)
        except BadRequest:
            return None

    def previous(self, page: OffsetPaging) -> OffsetPaging | None:
        """
        Retrieve the previous result set of a paging object.

        Parameters
        ----------
        page
            offset-based paging object

        Returns
        -------
        OffsetPaging
            previous result set
        """
        if self.is_async:
            return self._async_previous(page)

        if page.previous is None:
            return None

        previous_set = self._get_paging_result(page.previous)
        return type(page)(**previous_set)

    async def _async_previous(self, page: OffsetPaging) -> OffsetPaging | None:
        if page.previous is None:
            return None

        previous_set = await self._get_paging_result(page.previous)
        return type(page)(**previous_set)

    def all_pages(self, page: Paging) -> Generator[Paging, None, None]:
        """
        Retrieve all pages of a paging.

        Request and yield new (next) pages until the end of the paging.
        The paging that was given as an argument is yielded as the first result.

        Parameters
        ----------
        page
            paging object

        Returns
        -------
        Generator
            all pages within a paging
        """
        if self.is_async:
            return self._async_all_pages(page)
        return self._sync_all_pages(page)

    def _sync_all_pages(self, page: Paging):
        current: Paging | None = page
        while current is not None:
            yield current
            current = self.next(current)

    async def _async_all_pages(self, page: Paging):
        current: Paging | None = page
        while current is not None:
            yield current
            current = await self._async_next(current)

    def all_items(self, page: Paging) -> Generator[Model, None, None]:
        """
        Retrieve all items from all pages of a paging.

        Request and yield new (next) items until the end of the paging.
        The items in the paging that was given as an argument are yielded first.

        Parameters
        ----------
        page
            paging object

        Returns
        -------
        Generator
            all items within a paging
        """
        if self.is_async:
            return self._async_all_items(page)
        return self._sync_all_items(page)

    def _sync_all_items(self, paging: Paging):
        for page in self.all_pages(paging):
            yield from page.items

    async def _async_all_items(self, paging: Paging):
        async for page in self._async_all_pages(paging):
            for item in page.items:
                yield item



from __future__ import annotations

from tekore._model.base import Item
from tekore._model.member import Copyright, Image
from tekore._model.serialise import Model


class Author(Model):
    """Audiobook author."""

    name: str


class Narrator(Model):
    """Audiobook narrator."""

    name: str


class Audiobook(Item):
    """Audiobook base."""

    authors: list[Author]
    available_markets: list[str] | None = None
    copyrights: list[Copyright]
    description: str
    edition: str | None
    explicit: bool
    external_urls: dict
    html_description: str
    images: list[Image]
    is_externally_hosted: bool | None = None
    languages: list[str]
    media_type: str
    name: str
    narrators: list[Narrator]
    publisher: str
    total_chapters: int | None



from __future__ import annotations

from datetime import datetime

from .album import SimpleAlbum
from .artist import SimpleArtist
from .base import Item
from .member import Restrictions
from .paging import OffsetPaging
from .serialise import Model


class TrackLink(Item):
    """Relinked track."""

    external_urls: dict


class Track(Item):
    """Track base."""

    artists: list[SimpleArtist]
    available_markets: list[str] | None = None
    disc_number: int
    duration_ms: int
    explicit: bool
    external_urls: dict
    is_local: bool
    is_playable: bool | None = None
    linked_from: TrackLink | None = None
    name: str
    preview_url: str | None = None
    restrictions: Restrictions | None = None
    track_number: int


class SimpleTrack(Track):
    """
    Simplified track object.

    When market is specified, :attr:`available_markets` is not available.
    :attr:`is_playable` is not available when market is not specified.
    :attr:`restrictions` is available if restrictions have been placed on
    the track, making it unplayable.
    """


class FullTrack(Track):
    """
    Complete track object.

    When market is specified, :attr:`available_markets` is not available.
    :attr:`is_playable` is not available when market is not specified.
    :attr:`restrictions` is available if restrictions have been placed on
    the track, making it unplayable.
    """

    album: SimpleAlbum
    external_ids: dict
    popularity: int


class FullTrackPaging(OffsetPaging):
    """Paging of full tracks."""

    items: list[FullTrack]


class SimpleTrackPaging(OffsetPaging):
    """Paging of simplified tracks."""

    items: list[SimpleTrack]


class Tracks(Model):
    """Minimal representation of playlist tracks."""

    href: str
    total: int


class SavedTrack(Model):
    """Track saved to library."""

    added_at: datetime
    track: FullTrack


class SavedTrackPaging(OffsetPaging):
    """Paging of tracks in library."""

    items: list[SavedTrack]



from __future__ import annotations

from .context import Context
from .device import Device
from .episode import FullEpisode
from .local import LocalTrack
from .serialise import Model, StrEnum
from .track import FullTrack


class CurrentlyPlayingType(StrEnum):
    """Type of currently playing item."""

    ad = "ad"
    episode = "episode"
    track = "track"
    unknown = "unknown"


class RepeatState(StrEnum):
    """Playback repeat state."""

    off = "off"
    track = "track"
    context = "context"


class Disallows(Model):
    """Disallowed player actions."""

    interrupting_playback: bool = False
    pausing: bool = False
    resuming: bool = False
    seeking: bool = False
    skipping_next: bool = False
    skipping_prev: bool = False
    toggling_repeat_context: bool = False
    toggling_shuffle: bool = False
    toggling_repeat_track: bool = False
    transferring_playback: bool = False


class Actions(Model):
    """Player actions."""

    disallows: Disallows


PlaybackItem = FullTrack | LocalTrack | FullEpisode | None


class CurrentlyPlaying(Model):
    """
    Current playback.

    :attr:`context`, :attr:`progress_ms` and :attr:`item` may be ``None``
    e.g. during a private session.
    """

    actions: Actions
    currently_playing_type: CurrentlyPlayingType
    is_playing: bool
    timestamp: int
    context: Context | None
    progress_ms: int | None
    item: PlaybackItem


class CurrentlyPlayingContext(CurrentlyPlaying):
    """
    Extended current playback context.

    ``smart_shuffle`` is not documented in the Spotify API.
    """

    device: Device
    repeat_state: RepeatState
    shuffle_state: bool
    smart_shuffle: bool | None


class Queue(Model):
    """Playback queue."""

    currently_playing: PlaybackItem
    queue: list[PlaybackItem]



from __future__ import annotations

from collections.abc import Callable
from functools import wraps

from tekore._auth import Scope, scope
from tekore._client.base import SpotifyBase
from tekore._sender import Request, Response
from tekore._sender import send_and_process as _send_and_process

from .handle import handle_errors


def send_and_process(post_func: Callable) -> Callable:
    """
    Decorate a Spotify endpoint to send a request and process its content.

    Parameters
    ----------
    post_func
        function to call with response JSON content
    """

    def parse_response(request: Request, response: Response):
        handle_errors(request, response)
        return post_func(response.content)

    return _send_and_process(parse_response)


def maximise_limit(max_limit: int) -> Callable:
    """
    Decorate a function to maximise the value of a 'limit' argument.

    Parameters
    ----------
    max_limit
        maximum value of the limit
    """

    def decorator(function: Callable) -> Callable:
        varnames = function.__code__.co_varnames
        arg_pos = varnames.index("limit") - 1

        @wraps(function)
        def wrapper(self: SpotifyBase, *args, **kwargs):
            if self.max_limits_on and len(args) <= arg_pos:
                kwargs.setdefault("limit", max_limit)
            return function(self, *args, **kwargs)

        return wrapper

    return decorator


def _add_doc_section(doc: str, section: str) -> str:
    """Add section with correct indentation to docstring."""
    items = doc.split("\n", maxsplit=2)

    if len(items) == 1:
        empty = ""
        head = items[0]
        body = ""
    else:
        empty, head, body = items

    indent = (len(head) - len(head.lstrip(" "))) * " "

    section = indent + section.replace("\n", "\n" + indent)
    return f"{empty}\n{head}\n\n{section}\n{body}"


def scopes(
    required: list[scope] | None = None, optional: list[scope] | None = None
) -> Callable:
    """
    List the scopes that a call uses.

    Provides ``required_scopes``, ``optional_scopes``
    and their combination ``scopes``.
    Also modifies the docstring to include scope information.

    Parameters
    ----------
    required
        required scopes
    optional
        optional scopes
    """
    required = required or []
    optional = optional or []
    required_scope = Scope(*required)
    optional_scope = Scope(*optional)
    doc_msg = "\n".join(
        [
            "| Required :class:`scope`: " + (str(required_scope) or "none"),
            "| Optional :class:`scope`: " + (str(optional_scope) or "none"),
        ]
    )

    def decorator(function: Callable) -> Callable:
        function.required_scope = required_scope
        function.optional_scope = optional_scope
        function.scope = required_scope + optional_scope
        function.__doc__ = _add_doc_section(function.__doc__, doc_msg)
        return function

    return decorator



from __future__ import annotations

from tekore._model.base import Item
from tekore._model.member import Image, ReleaseDatePrecision, Restrictions, ResumePoint


class Chapter(Item):
    """Audiobook chapter base."""

    audio_preview_url: str | None
    available_markets: list[str] | None = None
    chapter_number: int
    description: str
    duration_ms: int
    explicit: bool
    external_urls: dict
    html_description: str
    images: list[Image]
    is_playable: bool | None = None
    languages: list[str]
    name: str
    release_date_precision: ReleaseDatePrecision
    release_date: str
    restrictions: Restrictions | None = None
    resume_point: ResumePoint | None = None



"""
Client for the Spotify Web API.

This module holds all public objects provided by Tekore
with the exception of response models, which are located in ``tekore.model``.

See online documentation at `RTD <http://tekore.rtfd.io>`_.
"""

from tekore import model

from ._auth import (
    AccessToken,
    Credentials,
    RefreshingCredentials,
    RefreshingToken,
    Scope,
    Token,
    UserAuth,
    gen_state,
    parse_code_from_url,
    parse_state_from_url,
    prompt_for_pkce_token,
    prompt_for_user_token,
    refresh_pkce_token,
    refresh_user_token,
    request_client_token,
    scope,
)
from ._client import Spotify, is_short_link
from ._config import (
    MissingConfigurationWarning,
    config_from_environment,
    config_from_file,
    config_to_file,
)
from ._convert import (
    ConversionError,
    IdentifierType,
    check_id,
    check_type,
    from_uri,
    from_url,
    to_uri,
    to_url,
)
from ._sender import (
    AsyncSender,
    BadGateway,
    BadRequest,
    CachingSender,
    Client,
    ClientError,
    ExtendingSender,
    Forbidden,
    HTTPError,
    InternalServerError,
    NotFound,
    Request,
    Response,
    RetryingSender,
    Sender,
    SenderConflictWarning,
    ServerError,
    ServiceUnavailable,
    SyncSender,
    TooManyRequests,
    Unauthorised,
)

__version__ = "6.1.0"

# Change the module of classes to hide module structure
# and fix Sphinx base class links
_classes = [
    Spotify,
    Credentials,
    Token,
    AccessToken,
    RefreshingCredentials,
    RefreshingToken,
    scope,
    Scope,
    UserAuth,
    HTTPError,
    ClientError,
    ServerError,
    BadRequest,
    Unauthorised,
    Forbidden,
    NotFound,
    TooManyRequests,
    InternalServerError,
    BadGateway,
    ServiceUnavailable,
    ConversionError,
    IdentifierType,
    Sender,
    SyncSender,
    AsyncSender,
    ExtendingSender,
    RetryingSender,
    CachingSender,
    SenderConflictWarning,
    Client,
    Request,
    Response,
    MissingConfigurationWarning,
]

for _cls in _classes:
    _cls.__module__ = "tekore"

client_id_var: str = "SPOTIFY_CLIENT_ID"
"""Configuration variable name for a client ID."""

client_secret_var: str = "SPOTIFY_CLIENT_SECRET"
"""Configuration variable name for a client secret."""

redirect_uri_var: str = "SPOTIFY_REDIRECT_URI"
"""Configuration variable name for a redirect URI."""

user_refresh_var: str = "SPOTIFY_USER_REFRESH"
"""Configuration variable name for a user refresh token."""



from enum import Enum, EnumMeta
from warnings import warn

from pydantic import BaseModel


class StrEnumMeta(EnumMeta):
    """
    Metaclass for StrEnum that provides case-insensitive get.

    This does not change values.
    """

    def __new__(mcs, cls, bases, classdict, **kwds):
        """Override `__new__` to make all keys lowercase."""
        enum_class = super().__new__(mcs, cls, bases, classdict, **kwds)
        copied_member_map = dict(enum_class._member_map_)
        enum_class._member_map_.clear()
        for k, v in copied_member_map.items():
            enum_class._member_map_[k.lower()] = v
        return enum_class

    def __getitem__(cls, name: str):
        # Ignore case on get item
        return super().__getitem__(name.lower())


class StrEnum(str, Enum, metaclass=StrEnumMeta):
    """
    Convert enumeration members to strings using their name.

    Ignores case when getting items. This does not change values.
    """

    @classmethod
    def _missing_(cls, value):
        return cls[value.lower()]

    def __str__(self) -> str:
        return self.name


class Model(BaseModel):
    """Response model base."""

    def __init__(self, **data) -> None:
        """"""  # noqa: D419
        super().__init__(**data)
        unknowns = set(data.keys()) - set(self.__dict__.keys())
        cls_name = self.__class__.__name__
        for arg in unknowns:
            msg = (
                f"{cls_name} contains unknown attribute: `{arg}`, which was discarded."
                " This warning may be safely ignored. Please consider upgrading Tekore."
            )
            warn(msg, UnknownModelAttributeWarning, stacklevel=5)


class UnknownModelAttributeWarning(RuntimeWarning):
    """The response model contains an unknown attribute."""



from __future__ import annotations

from tekore._auth import scope
from tekore._client.base import SpotifyBase
from tekore._client.decor import scopes, send_and_process
from tekore._client.process import nothing, single
from tekore._sender import Request
from tekore.model import FullPlaylist


class SpotifyPlaylistModify(SpotifyBase):
    """Playlist API endpoints for modifying playlists."""

    @scopes(
        [scope.playlist_modify_public, scope.ugc_image_upload],
        [scope.playlist_modify_private],
    )
    @send_and_process(nothing)
    def playlist_cover_image_upload(self, playlist_id: str, image: str) -> None:
        """
        Upload a custom playlist cover image.

        Parameters
        ----------
        playlist_id
            playlist ID
        image
            image data as a base64-encoded string
        """
        return (
            Request(
                method="PUT",
                url=f"playlists/{playlist_id}/images",
                headers={"Content-Type": "image/jpeg"},
                content=image,
            ),
            (),
        )

    @scopes([scope.playlist_modify_public], [scope.playlist_modify_private])
    @send_and_process(single(FullPlaylist))
    def playlist_create(
        self, user_id: str, name: str, public: bool = True, description: str = ""
    ) -> FullPlaylist:
        """
        Create a playlist.

        Parameters
        ----------
        user_id
            user ID
        name
            the name of the playlist
        public
            is the created playlist public
        description
            the description of the playlist
        """
        payload = {"name": name, "public": public, "description": description}
        return self._post(f"users/{user_id}/playlists", payload=payload)

    @scopes([scope.playlist_modify_public], [scope.playlist_modify_private])
    @send_and_process(nothing)
    def playlist_change_details(
        self,
        playlist_id: str,
        name: str | None = None,
        public: bool | None = None,
        collaborative: bool | None = None,
        description: str | None = None,
    ) -> None:
        """
        Change a playlist's details.

        Parameters
        ----------
        playlist_id
            playlist ID
        name
            name of the playlist
        public
            is the playlist public
        collaborative
            is the playlist collaborative
        description
            description of the playlist
        """
        payload = {
            "name": name,
            "public": public,
            "collaborative": collaborative,
            "description": description,
        }
        payload = {k: v for k, v in payload.items() if v is not None}
        return self._put("playlists/" + playlist_id, payload=payload)



from __future__ import annotations

from tekore._auth import scope
from tekore._client.base import SpotifyBase
from tekore._client.chunked import chunked, join_lists
from tekore._client.decor import maximise_limit, scopes, send_and_process
from tekore._client.process import model_list, single
from tekore.model import FullShow, SimpleEpisodePaging


class SpotifyShow(SpotifyBase):
    """Show API endpoints."""

    @scopes(optional=[scope.user_read_playback_position])
    @send_and_process(single(FullShow))
    def show(self, show_id: str, market: str | None = None) -> FullShow:
        """
        Get information for a show.

        The user-read-playback-position scope allows
        episode resume points to be returned.

        Parameters
        ----------
        show_id
            show ID
        market
            an ISO 3166-1 alpha-2 country code.
            If a user token is used to authenticate, the country associated
            with it overrides this parameter.
            If an application token is used and no market is specified,
            the show is considered unavailable.
        """
        return self._get("shows/" + show_id, market=market)

    @scopes(optional=[scope.user_read_playback_position])
    @chunked("show_ids", 1, 50, join_lists)
    @send_and_process(model_list(FullShow, "shows"))
    def shows(self, show_ids: list[str], market: str | None = None) -> list[FullShow]:
        """
        Get information for multiple shows.

        The user-read-playback-position scope allows
        episode resume points to be returned.

        Parameters
        ----------
        show_ids
            the show IDs, max 50 without chunking
        market
            an ISO 3166-1 alpha-2 country code.
            If a user token is used to authenticate, the country associated
            with it overrides this parameter.
            If an application token is used and no market is specified,
            the show is considered unavailable.
        """
        return self._get("shows", ids=",".join(show_ids), market=market)

    @scopes()
    @send_and_process(single(SimpleEpisodePaging))
    @maximise_limit(50)
    def show_episodes(
        self, show_id: str, market: str | None = None, limit: int = 20, offset: int = 0
    ) -> SimpleEpisodePaging:
        """
        Get episodes of a show.

        Parameters
        ----------
        show_id
            show ID
        market
            an ISO 3166-1 alpha-2 country code.
            If a user token is used to authenticate, the country associated
            with it overrides this parameter.
            If an application token is used and no market is specified,
            the show is considered unavailable.
        limit
            the number of items to return (1..50)
        offset
            the index of the first item to return
        """
        return self._get(
            f"shows/{show_id}/episodes", market=market, limit=limit, offset=offset
        )



from __future__ import annotations

from tekore._model.paging import OffsetPaging

from .base import Audiobook, Author, Narrator


class SimpleAudiobook(Audiobook):
    """
    Simplified audiobook.

    May contain :attr:`chapters`, but that is likely an error.
    """

    chapters: dict | None = None


class SimpleAudiobookPaging(OffsetPaging):
    """Paging of simplified audiobooks."""

    items: list[SimpleAudiobook]



from __future__ import annotations

from datetime import datetime

from .base import Item
from .member import Image, ReleaseDatePrecision, Restrictions, ResumePoint
from .paging import OffsetPaging
from .serialise import Model
from .show import SimpleShow


class Episode(Item):
    """
    Episode base.

    :attr:`language` is deprecated.
    """

    audio_preview_url: str | None
    description: str
    duration_ms: int
    explicit: bool
    external_urls: dict
    html_description: str
    images: list[Image]
    is_externally_hosted: bool
    is_playable: bool | None = None
    language: str | None = None
    languages: list[str]
    name: str
    release_date: str
    release_date_precision: ReleaseDatePrecision
    resume_point: ResumePoint | None = None


class SimpleEpisode(Episode):
    """Simplified episode object."""


class FullEpisode(Episode):
    """Complete episode object."""

    restrictions: Restrictions | None = None
    show: SimpleShow


class SimpleEpisodePaging(OffsetPaging):
    """Paging of simplified episodes."""

    items: list[SimpleEpisode]


class SavedEpisode(Model):
    """Episode saved to library."""

    added_at: datetime
    episode: FullEpisode


class SavedEpisodePaging(OffsetPaging):
    """Paging of episodes in library."""

    items: list[SavedEpisode]



from __future__ import annotations

from .base import Identifiable
from .serialise import StrEnum


class DeviceType(StrEnum):
    """Type of playback device."""

    Computer = "Computer"
    Tablet = "Tablet"
    Smartphone = "Smartphone"
    Speaker = "Speaker"
    TV = "TV"
    AVR = "AVR"
    STB = "STB"
    AudioDongle = "AudioDongle"
    GameConsole = "GameConsole"
    CastVideo = "CastVideo"
    CastAudio = "CastAudio"
    Automobile = "Automobile"
    Unknown = "Unknown"


class Device(Identifiable):
    """Playback device."""

    is_active: bool
    is_private_session: bool
    is_restricted: bool
    name: str
    type: DeviceType
    volume_percent: int | None
    supports_volume: bool



from tekore._model.paging import OffsetPaging

from .base import Chapter


class SimpleChapter(Chapter):
    """Simplified chapter."""


class SimpleChapterPaging(OffsetPaging):
    """Paging of simplified chapters."""

    items: list[SimpleChapter]



from .serialise import Model


class Identifiable(Model):
    """Object identified with a Spotify ID."""

    id: str


class Item(Identifiable):
    """Identifiable with additional fields."""

    href: str
    type: str
    uri: str



from itertools import product

from tekore.model import RecommendationAttribute

prefixes = {"min", "max", "target"}
attributes = {str(a) for a in RecommendationAttribute}
valid = {"_".join(i) for i in product(prefixes, attributes)}


def validate_attributes(candidates: dict) -> None:
    """
    Validate recommendation attributes.

    Parameters
    ----------
    candidates
        recommendation attributes to validate

    Raises
    ------
    ValueError
        if any attribute is not allowed
    """
    for name in candidates:
        if name not in valid:
            msg = f"Invalid attribute `{name}`!"
            raise ValueError(msg)



from __future__ import annotations

import time
from abc import ABC, abstractmethod

from tekore._auth.scope import Scope


class AccessToken(ABC):
    """Access token base class."""

    @property
    @abstractmethod
    def access_token(self) -> str:
        """
        Bearer token value.

        Used as the string representation of the instance.
        """
        raise NotImplementedError

    def __str__(self) -> str:
        """Bearer token value."""
        return self.access_token


class Token(AccessToken):
    """
    Expiring access token.

    Represents both client and user tokens.
    The refresh token of a client token is ``None``.
    """

    def __init__(self, token_info: dict, uses_pkce: bool) -> None:
        self._access_token = token_info["access_token"]
        self._token_type = token_info["token_type"]

        scope = token_info.get("scope", "")
        components = scope.split(" ")
        if components[0] == "":
            components = []
        self._scope = Scope(*components)

        self._refresh_token = token_info.get("refresh_token")
        self._expires_at = int(time.time()) + token_info["expires_in"]
        self._uses_pkce = uses_pkce

    def __repr__(self) -> str:
        options = [
            f"access_token={self.access_token!r}",
            f"refresh_token={self.refresh_token!r}",
            f"expires_at={self.expires_at!r}",
            f"scope={self.scope!r}",
        ]
        return type(self).__name__ + "(" + ", ".join(options) + ")"

    @property
    def access_token(self) -> str:
        """Bearer token value."""
        return self._access_token

    @property
    def refresh_token(self) -> str | None:
        """
        Refresh token for generating new access tokens.

        ``None`` if the token is an application token.
        """
        return self._refresh_token

    @property
    def token_type(self) -> str:
        """How the token may be used, always 'Bearer'."""
        return self._token_type

    @property
    def scope(self) -> Scope:
        """
        Privileges granted to the token.

        Empty :class:`Scope` if the token is an application token
        or a user token without any scopes.
        """
        return self._scope

    @property
    def expires_in(self) -> int:
        """Seconds until token expiration."""
        return self.expires_at - int(time.time())

    @property
    def expires_at(self) -> int:
        """When the token expires."""
        return self._expires_at

    @property
    def is_expiring(self) -> bool:
        """Determine whether token is about to expire."""
        return self.expires_in < 60  # noqa: PLR2004

    @property
    def uses_pkce(self) -> bool:
        """Proof key for code exchange used in authorisation."""
        return self._uses_pkce



from __future__ import annotations

from tekore._auth import scope
from tekore._client.base import SpotifyBase
from tekore._client.chunked import chunked, return_last
from tekore._client.decor import scopes, send_and_process
from tekore._client.process import nothing, top_item


class SpotifyPlaylistItems(SpotifyBase):
    """Playlist API endpoints for manipulating playlist items."""

    @scopes([scope.playlist_modify_public], [scope.playlist_modify_private])
    @chunked("uris", 2, 100, return_last, reverse="position", reverse_pos=3)
    @send_and_process(top_item("snapshot_id"))
    def playlist_add(
        self, playlist_id: str, uris: list[str], position: int | None = None
    ) -> str:
        """
        Add items.

        Parameters
        ----------
        playlist_id
            playlist ID
        uris
            list of URIs, max 100 without chunking
        position
            index to insert the items in

        Returns
        -------
        str
            snapshot ID for the playlist
        """
        return self._post(
            f"playlists/{playlist_id}/tracks", payload={"uris": uris}, position=position
        )

    @scopes([scope.playlist_modify_public], [scope.playlist_modify_private])
    @send_and_process(nothing)
    def playlist_clear(self, playlist_id: str) -> None:
        """
        Remove all items.

        Parameters
        ----------
        playlist_id
            playlist ID
        """
        return self._put(f"playlists/{playlist_id}/tracks", payload={"uris": []})

    @scopes([scope.playlist_modify_public], [scope.playlist_modify_private])
    @send_and_process(nothing)
    def playlist_replace(self, playlist_id: str, uris: list[str]) -> None:
        """
        Replace all items.

        Parameters
        ----------
        playlist_id
            playlist ID
        uris
            list of URIs, max 100
        """
        return self._put(f"playlists/{playlist_id}/tracks", payload={"uris": uris})

    @scopes([scope.playlist_modify_public], [scope.playlist_modify_private])
    @send_and_process(top_item("snapshot_id"))
    def playlist_reorder(
        self,
        playlist_id: str,
        range_start: int,
        insert_before: int,
        range_length: int = 1,
        snapshot_id: str | None = None,
    ) -> str:
        """
        Reorder items.

        Parameters
        ----------
        playlist_id
            playlist ID
        range_start
            position of the first item to be reordered
        range_length
            number of items to be reordered
        insert_before
            position where the items should be inserted
        snapshot_id
            snapshot ID for the playlist

        Returns
        -------
        str
            snapshot ID for the playlist
        """
        payload: dict[str, int | str] = {
            "range_start": range_start,
            "range_length": range_length,
            "insert_before": insert_before,
        }
        if snapshot_id:
            payload["snapshot_id"] = snapshot_id
        return self._put(f"playlists/{playlist_id}/tracks", payload=payload)

    def _generic_playlist_remove(
        self, playlist_id: str, payload: dict, snapshot_id: str | None = None
    ) -> str:
        if snapshot_id:
            payload["snapshot_id"] = snapshot_id
        return self._delete(f"playlists/{playlist_id}/tracks", payload=payload)

    @scopes([scope.playlist_modify_public], [scope.playlist_modify_private])
    @chunked("uris", 2, 100, return_last, chain="snapshot_id", chain_pos=3)
    @send_and_process(top_item("snapshot_id"))
    def playlist_remove(
        self, playlist_id: str, uris: list[str], snapshot_id: str | None = None
    ) -> str:
        """
        Remove items by URI.

        Removes all occurrences of the specified items.
        Note that when chunked, ``snapshot_id`` is not updated between requests.

        Parameters
        ----------
        playlist_id
            playlist ID
        uris
            list of URIs, max 100 without chunking
        snapshot_id
            snapshot ID for the playlist

        Returns
        -------
        str
            snapshot ID for the playlist
        """
        items = [{"uri": uri} for uri in uris]
        return self._generic_playlist_remove(
            playlist_id, {"tracks": items}, snapshot_id
        )



from __future__ import annotations

from .serialise import Model


class TimeInterval(Model):
    """
    Generic representation of an interval.

    Attributes are sometimes not available.
    """

    duration: float
    start: float | None = None
    confidence: float | None = None


class Section(Model):
    """
    Analysis of a track's section.

    Attributes are sometimes not available.
    """

    duration: float
    loudness: float
    tempo: float
    tempo_confidence: float
    key_confidence: float
    mode_confidence: float
    time_signature: int
    time_signature_confidence: float
    confidence: float | None = None
    mode: int | None = None
    key: int | None = None
    start: float | None = None


class Segment(Model):
    """
    Analysis of a track's segment.

    Attributes are sometimes not available.
    """

    duration: float
    loudness_start: float
    loudness_max: float
    pitches: list[float]
    timbre: list[float]
    confidence: float | None = None
    loudness_end: float | None = None
    loudness_max_time: float | None = None
    start: float | None = None


class AudioAnalysis(Model):
    """
    Track audio analysis.

    See the Web API
    `documentation <https://developer.spotify.com/documentation/web-api/\
    reference/tracks/get-audio-analysis/>`_
    for more details.
    """

    bars: list[TimeInterval]
    beats: list[TimeInterval]
    sections: list[Section]
    segments: list[Segment]
    tatums: list[TimeInterval]
    meta: dict
    track: dict



from tekore._auth import scope
from tekore._client.base import SpotifyBase
from tekore._client.decor import maximise_limit, scopes, send_and_process
from tekore._client.process import single
from tekore.model import FullArtistOffsetPaging, FullTrackPaging


class SpotifyPersonalisation(SpotifyBase):
    """Personalisation API endpoints."""

    @scopes([scope.user_top_read])
    @send_and_process(single(FullArtistOffsetPaging))
    @maximise_limit(50)
    def current_user_top_artists(
        self, time_range: str = "medium_term", limit: int = 20, offset: int = 0
    ) -> FullArtistOffsetPaging:
        """
        Get the current user's top artists.

        Parameters
        ----------
        time_range
            Over what time frame are the affinities computed.
            Valid-values: short_term, medium_term, long_term
        limit
            the number of items to return (1..50)
        offset
            the index of the first item to return
        """
        return self._get(
            "me/top/artists", time_range=time_range, limit=limit, offset=offset
        )

    @scopes([scope.user_top_read])
    @send_and_process(single(FullTrackPaging))
    @maximise_limit(50)
    def current_user_top_tracks(
        self, time_range: str = "medium_term", limit: int = 20, offset: int = 0
    ) -> FullTrackPaging:
        """
        Get the current user's top tracks.

        Parameters
        ----------
        time_range
            Over what time frame are the affinities computed.
            Valid-values: short_term, medium_term, long_term
        limit
            the number of items to return (1..50)
        offset
            the index of the first item to return
        """
        return self._get(
            "me/top/tracks", time_range=time_range, limit=limit, offset=offset
        )



from __future__ import annotations

from collections.abc import Iterable
from configparser import ConfigParser
from os import environ
from pathlib import Path
from warnings import warn

import tekore as tk


class MissingConfigurationWarning(RuntimeWarning):
    """Missing value read from configuration."""


def _read_configuration(conf: dict, *, return_refresh: bool = False) -> tuple:
    """
    Read variables from dictionary.

    Parameters
    ----------
    conf
        dictionary to read from
    return_refresh
        return user refresh token

    Returns
    -------
    tuple
        (client ID, client secret, redirect URI), None if not found.
        If return_refresh is True, also return user refresh token.
    """
    variables = [tk.client_id_var, tk.client_secret_var, tk.redirect_uri_var]

    if return_refresh:
        variables.append(tk.user_refresh_var)

    values = tuple(conf.get(var) for var in variables)

    for var, val in zip(variables, values, strict=True):
        if val is None:
            warn(
                f"`{var}` missing! None returned instead.",
                MissingConfigurationWarning,
                stacklevel=3,
            )

    return values


def config_from_environment(return_refresh: bool = False) -> tuple:
    """
    Read configuration from environment variables.

    Parameters
    ----------
    return_refresh
        return user refresh token

    Returns
    -------
    tuple
        (client ID, client secret, redirect URI), None if not found.
        If return_refresh is True, also return user refresh token.

    Raises
    ------
    MissingConfigurationWarning
        when a missing value is encountered

    Examples
    --------
    .. code:: python

        client_id, client_secret, redirect_uri = tk.config_from_environment()

        conf = tk.config_from_environment(return_refresh=True)
        client_id, client_secret, redirect_uri, user_refresh = conf
    """
    return _read_configuration(environ, return_refresh=return_refresh)


def _read_configfile(file_path: str, *, force: bool = True) -> ConfigParser:
    """
    Read configuration from INI file.

    Parameters
    ----------
    file_path
        path of the configuration file
    force
        force reading of the file, fail if not found
    """
    c = ConfigParser()
    c.optionxform = str

    if force:
        with Path(file_path).open() as f:
            c.read_file(f)
    else:
        c.read(file_path)

    return c


def config_from_file(
    file_path: str, section: str = "DEFAULT", return_refresh: bool = False
) -> tuple:
    """
    Read configuration from a config file.

    The configuration must be in INI format
    as accepted by :class:`configparser.ConfigParser`.

    Parameters
    ----------
    file_path
        path of the file containing the credential variables
    section
        name of the section to read variables from
    return_refresh
        return user refresh token

    Returns
    -------
    tuple
        (client ID, client secret, redirect URI), None if not found.
        If return_refresh is True, also return user refresh token.

    Raises
    ------
    MissingConfigurationWarning
        when a missing value is encountered

    Examples
    --------
    .. code:: python

        client_id, client_secret, redirect_uri = tk.config_from_file(filename)

        conf = tk.config_from_file(filename, return_refresh=True)
        client_id, client_secret, redirect_uri, user_refresh = conf
    """
    c = _read_configfile(file_path)
    return _read_configuration(c[section], return_refresh=return_refresh)


def config_to_file(
    file_path: str, values: Iterable | dict, section: str = "DEFAULT"
) -> None:
    """
    Write configuration to a config file.

    Existing configuration is preserved if it's not in conflict.

    Parameters
    ----------
    file_path
        path of the configuration file
    values
        configuration values to write, dict or iterable, see below for examples
    section
        name of the section to write to

    Examples
    --------
    Configuration can be written in different ways.
    Pass in an iterable to use the preset variable names.
    The values should be ordered as returned when reading configuration:
    ``client_id, client_secret, redirect_uri, user_refresh``.

    .. code:: python

        conf = (client_id, client_secret, redirect_uri, user_refresh)
        config_to_file(filename, conf)

    A shorter iterable or one containing ``None`` values may be passed.
    Items missing from the end are ignored and ``None`` values are discarded.

    .. code:: python

        # Write partial information
        config_to_file(filename, (client_id, client_secret))

        # Fill the missing configuration
        conf = (None, None, redirect_uri, user_refresh)
        config_to_file(filename, conf)

    A dictionary is also accepted.
    In this case the keys are used instead of preset variable names.

    .. code:: python

        config_to_file(filename, {'REFRESH_TOKEN': refresh_token})
    """
    if isinstance(values, dict):
        val_dict = values
    else:
        names = (
            tk.client_id_var,
            tk.client_secret_var,
            tk.redirect_uri_var,
            tk.user_refresh_var,
        )
        val_dict = {
            name: value
            for name, value in zip(names, values, strict=False)
            if value is not None
        }

    c = _read_configfile(file_path, force=False)

    if section not in c:
        c[section] = {}

    c[section].update(val_dict)

    with Path(file_path).open("w") as f:
        c.write(f)



from collections.abc import Generator
from contextlib import contextmanager

from .api import (
    SpotifyAlbum,
    SpotifyArtist,
    SpotifyAudiobook,
    SpotifyBrowse,
    SpotifyChapter,
    SpotifyEpisode,
    SpotifyFollow,
    SpotifyLibrary,
    SpotifyMarkets,
    SpotifyPersonalisation,
    SpotifyPlayer,
    SpotifyPlaylist,
    SpotifySearch,
    SpotifyShow,
    SpotifyTrack,
    SpotifyUser,
)
from .paging import SpotifyPaging
from .short_link import SpotifyShortLink


class Spotify(
    SpotifyAlbum,
    SpotifyArtist,
    SpotifyAudiobook,
    SpotifyBrowse,
    SpotifyChapter,
    SpotifyEpisode,
    SpotifyFollow,
    SpotifyLibrary,
    SpotifyMarkets,
    SpotifyPersonalisation,
    SpotifyPlayer,
    SpotifyPlaylist,
    SpotifySearch,
    SpotifyShow,
    SpotifyTrack,
    SpotifyUser,
    SpotifyPaging,
    SpotifyShortLink,
):
    """
    Bases: :class:`tekore.Client`.

    Client to Web API endpoints.

    Parameters
    ----------
    token
        bearer token for requests
    sender
        request sender
    asynchronous
        synchronicity requirement
    max_limits_on
        use maximum limits in paging calls, overrided by endpoint arguments
    chunked_on
        use chunking when requesting lists of resources

    Attributes
    ----------
    token
        bearer token for requests
    sender
        underlying sender
    max_limits_on
        use maximum limits in paging calls, overrided by endpoint arguments
    chunked_on
        use chunking when requesting lists of resources
    """

    @contextmanager
    def token_as(self, token) -> Generator["Spotify", None, None]:
        """
        Use a different token with requests. Context manager, async safe.

        Parameters
        ----------
        token
            access token

        Returns
        -------
        Generator[Spotify, None, None]
            self as context

        Examples
        --------
        .. code:: python

            spotify = Spotify()
            with spotify.token_as(token):
                album = spotify.album(album_id)

            spotify = Spotify(app_token)
            with spotify.token_as(user_token):
                user = spotify.current_user()
        """
        cv_token = self._token_cv.set(token)
        yield self
        self._token_cv.reset(cv_token)

    @contextmanager
    def max_limits(self, on: bool = True) -> Generator["Spotify", None, None]:
        """
        Toggle using maximum limits in paging calls. Context manager, async safe.

        Parameters
        ----------
        on
            enable or disable using maximum limits

        Returns
        -------
        Generator[Spotify, None, None]
            self as context

        Examples
        --------
        .. code:: python

            spotify = Spotify(token)
            with spotify.max_limits(True):
                tracks, = spotify.search('piano')

            spotify = Spotify(token, max_limits_on=True)
            with spotify.max_limits(False):
                tracks, = spotify.search('piano')
        """
        cv_token = self._max_limits_on_cv.set(on)
        yield self
        self._max_limits_on_cv.reset(cv_token)

    @contextmanager
    def chunked(self, on: bool = True) -> Generator["Spotify", None, None]:
        """
        Toggle chunking lists of resources. Context manager, async safe.

        Parameters
        ----------
        on
            enable or disable chunking

        Returns
        -------
        Generator[Spotify, None, None]
            self as context

        Examples
        --------
        .. code:: python

            spotify = Spotify(token)
            with spotify.chunked(True):
                tracks = spotify.tracks(many_ids)

            spotify = Spotify(token, chunked_on=True)
            with spotify.chunked(False):
                tracks = spotify.search(many_ids[:50])
        """
        cv_token = self._chunked_on_cv.set(on)
        yield self
        self._chunked_on_cv.reset(cv_token)



from __future__ import annotations

from datetime import datetime

from tekore._model.album.base import Album
from tekore._model.member import Copyright
from tekore._model.paging import OffsetPaging
from tekore._model.serialise import Model
from tekore._model.track import SimpleTrackPaging


class FullAlbum(Album):
    """
    Complete album object.

    :attr:`available_markets` is available when market is not specified.

    The presence of :attr:`is_playable` is undocumented
    and it appears to only be ``True`` when it is present.
    """

    copyrights: list[Copyright]
    external_ids: dict
    genres: list[str]
    label: str | None
    popularity: int
    tracks: SimpleTrackPaging


class SavedAlbum(Model):
    """Album saved to library."""

    added_at: datetime
    album: FullAlbum


class SavedAlbumPaging(OffsetPaging):
    """Paging of albums in library."""

    items: list[SavedAlbum]



"""Response model definitions for client."""

from tekore._model import (
    Actions,
    Album,
    AlbumGroup,
    AlbumType,
    Artist,
    AudioAnalysis,
    Audiobook,
    AudioFeatures,
    Author,
    Category,
    CategoryPaging,
    Chapter,
    Context,
    ContextType,
    Copyright,
    CurrentlyPlaying,
    CurrentlyPlayingContext,
    CurrentlyPlayingType,
    Cursor,
    CursorPaging,
    Device,
    DeviceType,
    Disallows,
    Episode,
    ExplicitContent,
    Followers,
    FullAlbum,
    FullArtist,
    FullArtistCursorPaging,
    FullArtistOffsetPaging,
    FullAudiobook,
    FullChapter,
    FullEpisode,
    FullPlaylist,
    FullPlaylistEpisode,
    FullPlaylistTrack,
    FullShow,
    FullTrack,
    FullTrackPaging,
    Identifiable,
    Image,
    Item,
    LocalAlbum,
    LocalArtist,
    LocalItem,
    LocalPlaylistTrack,
    LocalTrack,
    Model,
    Narrator,
    OffsetPaging,
    Paging,
    PlayerErrorReason,
    PlayHistory,
    PlayHistoryCursor,
    PlayHistoryPaging,
    Playlist,
    PlaylistTrack,
    PlaylistTrackPaging,
    PrivateUser,
    PublicUser,
    Queue,
    RecommendationAttribute,
    Recommendations,
    RecommendationSeed,
    ReleaseDatePrecision,
    RepeatState,
    Restrictions,
    ResumePoint,
    SavedAlbum,
    SavedAlbumPaging,
    SavedEpisode,
    SavedEpisodePaging,
    SavedShow,
    SavedShowPaging,
    SavedTrack,
    SavedTrackPaging,
    Section,
    Segment,
    Show,
    SimpleAlbum,
    SimpleAlbumPaging,
    SimpleArtist,
    SimpleAudiobook,
    SimpleAudiobookPaging,
    SimpleChapter,
    SimpleChapterPaging,
    SimpleEpisode,
    SimpleEpisodePaging,
    SimplePlaylist,
    SimplePlaylistPaging,
    SimpleShow,
    SimpleShowPaging,
    SimpleTrack,
    SimpleTrackPaging,
    StrEnum,
    TimeInterval,
    Track,
    TrackLink,
    Tracks,
    UnknownModelAttributeWarning,
    User,
)

# Change the module of classes to hide module structure
# and fix Sphinx base class links
_classes = [
    AlbumType,
    Album,
    AlbumGroup,
    SimpleAlbum,
    SimpleAlbumPaging,
    FullAlbum,
    SavedAlbum,
    SavedAlbumPaging,
    Artist,
    SimpleArtist,
    FullArtist,
    FullArtistCursorPaging,
    FullArtistOffsetPaging,
    AudioAnalysis,
    TimeInterval,
    Section,
    Segment,
    AudioFeatures,
    Author,
    Narrator,
    Audiobook,
    SimpleAudiobook,
    SimpleAudiobookPaging,
    FullAudiobook,
    Identifiable,
    Item,
    Category,
    CategoryPaging,
    Chapter,
    SimpleChapter,
    SimpleChapterPaging,
    FullChapter,
    ContextType,
    Context,
    CurrentlyPlayingType,
    CurrentlyPlayingContext,
    CurrentlyPlaying,
    RepeatState,
    Disallows,
    Actions,
    Queue,
    Device,
    DeviceType,
    ResumePoint,
    Episode,
    SimpleEpisode,
    SimpleEpisodePaging,
    FullEpisode,
    SavedEpisode,
    SavedEpisodePaging,
    PlayerErrorReason,
    LocalItem,
    LocalAlbum,
    LocalArtist,
    LocalTrack,
    ReleaseDatePrecision,
    Copyright,
    Followers,
    Image,
    Restrictions,
    Paging,
    OffsetPaging,
    Cursor,
    CursorPaging,
    PlayHistory,
    PlayHistoryCursor,
    PlayHistoryPaging,
    PlaylistTrack,
    PlaylistTrackPaging,
    Playlist,
    SimplePlaylist,
    FullPlaylist,
    SimplePlaylistPaging,
    FullPlaylistTrack,
    FullPlaylistEpisode,
    LocalPlaylistTrack,
    Recommendations,
    RecommendationSeed,
    RecommendationAttribute,
    Show,
    SimpleShow,
    SimpleShowPaging,
    SavedShow,
    SavedShowPaging,
    FullShow,
    TrackLink,
    Track,
    Tracks,
    SimpleTrack,
    SavedTrack,
    FullTrack,
    SimpleTrackPaging,
    SavedTrackPaging,
    FullTrackPaging,
    ExplicitContent,
    UnknownModelAttributeWarning,
    User,
    PrivateUser,
    PublicUser,
    Model,
    StrEnum,
]

for _cls in _classes:
    _cls.__module__ = "tekore.model"



from __future__ import annotations

from datetime import datetime

from .context import Context
from .paging import Cursor, CursorPaging
from .serialise import Model
from .track import FullTrack


class PlayHistory(Model):
    """
    Previously played track.

    Context is supposedly sometimes available.
    """

    track: FullTrack
    played_at: datetime
    context: Context | None


class PlayHistoryCursor(Cursor):
    """Cursor to play history."""

    before: str


class PlayHistoryPaging(CursorPaging):
    """
    Paging to play history.

    Cursors are not available when paging is exhausted.
    """

    items: list[PlayHistory]
    cursors: PlayHistoryCursor | None



from .items import SpotifyPlaylistItems
from .modify import SpotifyPlaylistModify
from .view import SpotifyPlaylistView


class SpotifyPlaylist(SpotifyPlaylistView, SpotifyPlaylistModify, SpotifyPlaylistItems):
    """All playlist API endpoints."""



from __future__ import annotations

from tekore._model.album.base import Album, AlbumType
from tekore._model.paging import OffsetPaging
from tekore._model.serialise import StrEnum


class AlbumGroup(StrEnum):
    """Relationship between artist and album."""

    album = "album"
    appears_on = "appears_on"
    compilation = "compilation"
    single = "single"


class SimpleAlbum(Album):
    """
    Simplified album object.

    :attr:`album_group` is available when getting an artist's albums.
    :attr:`available_markets` is available when market is not specified.

    The presence of :attr:`is_playable` is undocumented
    and it appears to only be ``True`` when it is present.
    """

    album_group: AlbumGroup | None = None


class SimpleAlbumPaging(OffsetPaging):
    """Paging containing simplified albums."""

    items: list[SimpleAlbum]



from __future__ import annotations

from tekore._client.base import SpotifyBase
from tekore._client.chunked import chunked, join_lists
from tekore._client.decor import maximise_limit, scopes, send_and_process
from tekore._client.process import model_list, single
from tekore.model import AlbumGroup, FullArtist, FullTrack, SimpleAlbumPaging


class SpotifyArtist(SpotifyBase):
    """Artist API endpoints."""

    @scopes()
    @send_and_process(single(FullArtist))
    def artist(self, artist_id: str) -> FullArtist:
        """
        Get information for an artist.

        Parameters
        ----------
        artist_id
            artist ID
        """
        return self._get("artists/" + artist_id)

    @scopes()
    @chunked("artist_ids", 1, 50, join_lists)
    @send_and_process(model_list(FullArtist, "artists"))
    def artists(self, artist_ids: list[str]) -> list[FullArtist]:
        """
        Get information for multiple artists.

        Parameters
        ----------
        artist_ids
            list of artist IDs, max 50 without chunking
        """
        return self._get("artists", ids=",".join(artist_ids))

    @scopes()
    @send_and_process(single(SimpleAlbumPaging))
    @maximise_limit(50)
    def artist_albums(
        self,
        artist_id: str,
        include_groups: list[str | AlbumGroup] | None = None,
        market: str | None = None,
        limit: int = 20,
        offset: int = 0,
    ) -> SimpleAlbumPaging:
        """
        Get an artist's albums.

        Parameters
        ----------
        artist_id
            the artist ID
        include_groups
            album groups to include in the response
        market
            an ISO 3166-1 alpha-2 country code or 'from_token'
        limit
            the number of items to return (1..50)
        offset
            the index of the first item to return
        """
        if include_groups is not None:
            include_groups = ",".join(str(g) for g in include_groups)
        return self._get(
            f"artists/{artist_id}/albums",
            include_groups=include_groups,
            market=market,
            limit=limit,
            offset=offset,
        )

    @scopes()
    @send_and_process(model_list(FullTrack, "tracks"))
    def artist_top_tracks(self, artist_id: str, market: str) -> list[FullTrack]:
        """
        Get an artist's top 10 tracks by country.

        Parameters
        ----------
        artist_id
            the artist ID
        market
            an ISO 3166-1 alpha-2 country code or 'from_token'
        """
        return self._get(f"artists/{artist_id}/top-tracks", country=market)

    @scopes()
    @send_and_process(model_list(FullArtist, "artists"))
    def artist_related_artists(self, artist_id: str) -> list[FullArtist]:
        """
        Get artists similar to an identified artist.

        .. warning::

            This endpoint is unavailable to new third-party applications (:issue:`331`)

        Similarity is based on analysis of
        the Spotify community's listening history.

        Parameters
        ----------
        artist_id
            artist ID
        """
        return self._get(f"artists/{artist_id}/related-artists")



from __future__ import annotations

from tekore._client.base import SpotifyBase
from tekore._client.decor import maximise_limit, scopes, send_and_process
from tekore._client.process import multiple, single, top_item
from tekore.model import (
    Category,
    CategoryPaging,
    Recommendations,
    SimpleAlbumPaging,
    SimplePlaylistPaging,
)

from .validate import validate_attributes


class SpotifyBrowse(SpotifyBase):
    """Browse API endpoints."""

    @scopes()
    @send_and_process(
        multiple(
            top_item("message"), single(SimplePlaylistPaging, from_item="playlists")
        )
    )
    @maximise_limit(50)
    def featured_playlists(
        self,
        country: str | None = None,
        locale: str | None = None,
        timestamp: str | None = None,
        limit: int = 20,
        offset: int = 0,
    ) -> tuple[str, SimplePlaylistPaging]:
        """
        Get a list of Spotify featured playlists.

        .. warning::

            This endpoint is unavailable to new third-party applications (:issue:`331`)

        Parameters
        ----------
        country
            an ISO 3166-1 alpha-2 country code
        locale
            the desired language, consisting of a lowercase ISO 639 language code
            and an uppercase ISO 3166-1 alpha-2 country code joined by an underscore
        timestamp
            Timestamp in ISO 8601 format: yyyy-MM-ddTHH:mm:ss.
            Used to specify the user's local time to get results tailored for
            that specific date and time in the day.
        limit
            the number of items to return (1..50)
        offset
            the index of the first item to return

        Returns
        -------
        tuple[str, SimplePlaylistPaging]
            message and playlists
        """
        return self._get(
            "browse/featured-playlists",
            locale=locale,
            country=country,
            timestamp=timestamp,
            limit=limit,
            offset=offset,
        )

    @scopes()
    @send_and_process(single(SimpleAlbumPaging, from_item="albums"))
    @maximise_limit(50)
    def new_releases(
        self, country: str | None = None, limit: int = 20, offset: int = 0
    ) -> SimpleAlbumPaging:
        """
        Get a list of new album releases featured in Spotify.

        Parameters
        ----------
        country
            an ISO 3166-1 alpha-2 country code
        limit
            the number of items to return (1..50)
        offset
            the index of the first item to return
        """
        return self._get(
            "browse/new-releases", country=country, limit=limit, offset=offset
        )

    @scopes()
    @send_and_process(single(CategoryPaging, from_item="categories"))
    @maximise_limit(50)
    def categories(
        self,
        country: str | None = None,
        locale: str | None = None,
        limit: int = 20,
        offset: int = 0,
    ) -> CategoryPaging:
        """
        Get a list of categories used to tag items in Spotify.

        Parameters
        ----------
        country
            an ISO 3166-1 alpha-2 country code
        locale
            the desired language, consisting of a lowercase ISO 639 language code
            and an uppercase ISO 3166-1 alpha-2 country code joined by an underscore
        limit
            the number of items to return (1..50)
        offset
            the index of the first item to return
        """
        return self._get(
            "browse/categories",
            country=country,
            locale=locale,
            limit=limit,
            offset=offset,
        )

    @scopes()
    @send_and_process(single(Category))
    def category(
        self, category_id: str, country: str | None = None, locale: str | None = None
    ) -> Category:
        """
        Get a single category used to tag items in Spotify.

        Parameters
        ----------
        category_id
            category ID
        country
            an ISO 3166-1 alpha-2 country code
        locale
            the desired language, consisting of a lowercase ISO 639 language code
            and an uppercase ISO 3166-1 alpha-2 country code joined by an underscore
        """
        return self._get(
            "browse/categories/" + category_id, country=country, locale=locale
        )

    @scopes()
    @send_and_process(single(SimplePlaylistPaging, from_item="playlists"))
    @maximise_limit(50)
    def category_playlists(
        self,
        category_id: str,
        country: str | None = None,
        limit: int = 20,
        offset: int = 0,
    ) -> SimplePlaylistPaging:
        """
        Get a list of Spotify playlists tagged with a particular category.

        .. warning::

            This endpoint is unavailable to new third-party applications (:issue:`331`)

        Parameters
        ----------
        category_id
            category ID
        country
            an ISO 3166-1 alpha-2 country code
        limit
            the number of items to return (1..50)
        offset
            the index of the first item to return
        """
        return self._get(
            f"browse/categories/{category_id}/playlists",
            country=country,
            limit=limit,
            offset=offset,
        )

    @scopes()
    @send_and_process(single(Recommendations))
    @maximise_limit(100)
    def recommendations(
        self,
        artist_ids: list[str] | None = None,
        genres: list[str] | None = None,
        track_ids: list[str] | None = None,
        limit: int = 20,
        market: str | None = None,
        **attributes,
    ) -> Recommendations:
        """
        Get a list of recommended tracks for seeds.

        .. warning::

            This endpoint is unavailable to new third-party applications (:issue:`331`)

        .. warning::
            The total number of seeds provided in ``artist_ids``, ``genres``
            and ``track_ids`` must be at least 1 and at most 5.

        Parameters
        ----------
        artist_ids
            list of seed artist IDs
        genres
            list of seed genre names
        track_ids
            list of seed track IDs
        limit
            the number of items to return (1..100)
        market
            an ISO 3166-1 alpha-2 country code or 'from_token'
        attributes
            min/max/target_<attribute> - For all tuneable track attributes see
            :class:`model.RecommendationAttribute`,
            these values provide filters and targeting on results.

        Raises
        ------
        ValueError
            if any attribute is not allowed
        """
        params: dict[str, int | str] = {"limit": limit}
        if artist_ids is not None:
            params["seed_artists"] = ",".join(artist_ids)
        if genres is not None:
            params["seed_genres"] = ",".join(genres)
        if track_ids is not None:
            params["seed_tracks"] = ",".join(track_ids)
        if market is not None:
            params["market"] = market

        validate_attributes(attributes)
        params.update(attributes)

        return self._get("recommendations", **params)

    @scopes()
    @send_and_process(top_item("genres"))
    def recommendation_genre_seeds(self) -> list[str]:
        """Get a list of available genre seeds."""
        return self._get("recommendations/available-genre-seeds")



from __future__ import annotations

from .base import Item
from .member import Followers, Image
from .serialise import Model


class ExplicitContent(Model):
    """Explicit content filter of a user."""

    filter_enabled: bool
    filter_locked: bool


class User(Item):
    """
    User base.

    :attr:`display_name`, :attr:`followers` and :attr:`images`
    may not be available.
    """

    external_urls: dict
    display_name: str | None = None
    followers: Followers | None = None
    images: list[Image] | None = None


class PrivateUser(User):
    """
    User with private information.

    :attr:`country`, :attr:`explicit_content` and :attr:`product`
    require the ``user-read-private`` scope.
    :attr:`email` requires the ``user-read-email`` scope.
    :attr:`birthday` is unavailable unless the now-invalid
    ``user-read-birthdate`` scope was granted to the token.
    """

    country: str | None = None
    email: str | None = None
    explicit_content: ExplicitContent | None = None
    product: str | None = None
    birthday: str | None = None


class PublicUser(User):
    """User as viewable by anyone."""



from datetime import datetime

from tekore._model.paging import OffsetPaging
from tekore._model.serialise import Model
from tekore._model.show.base import Show


class SimpleShow(Show):
    """
    Simplified show object.

    :attr:`total_episodes` is undocumented by Spotify,
    so it might be missing or removed in a future version.
    """


class SimpleShowPaging(OffsetPaging):
    """Paging of simplified shows."""

    items: list[SimpleShow]


class SavedShow(Model):
    """Show saved in library."""

    added_at: datetime
    show: SimpleShow


class SavedShowPaging(OffsetPaging):
    """Paging of shows in library."""

    items: list[SavedShow]



from __future__ import annotations

from tekore._client.base import SpotifyBase
from tekore._client.chunked import chunked, join_lists
from tekore._client.decor import maximise_limit, scopes, send_and_process
from tekore._client.process import model_list, single
from tekore.model import FullAlbum, SimpleTrackPaging


class SpotifyAlbum(SpotifyBase):
    """Album API endpoints."""

    @scopes()
    @send_and_process(single(FullAlbum))
    def album(self, album_id: str, market: str | None = None) -> FullAlbum:
        """
        Get an album.

        Parameters
        ----------
        album_id
            album ID
        market
            an ISO 3166-1 alpha-2 country code or 'from_token'
        """
        return self._get("albums/" + album_id, market=market)

    @scopes()
    @send_and_process(single(SimpleTrackPaging))
    @maximise_limit(50)
    def album_tracks(
        self, album_id: str, market: str | None = None, limit: int = 20, offset: int = 0
    ) -> SimpleTrackPaging:
        """
        Get tracks on album.

        Parameters
        ----------
        album_id
            album ID
        market
            an ISO 3166-1 alpha-2 country code or 'from_token'
        limit
            the number of items to return (1..50)
        offset
            the index of the first item to return
        """
        return self._get(
            f"albums/{album_id}/tracks", market=market, limit=limit, offset=offset
        )

    @scopes()
    @chunked("album_ids", 1, 20, join_lists)
    @send_and_process(model_list(FullAlbum, "albums"))
    def albums(
        self, album_ids: list[str], market: str | None = None
    ) -> list[FullAlbum]:
        """
        Get multiple albums.

        Parameters
        ----------
        album_ids
            list of album IDs, max 20 without chunking
        market
            an ISO 3166-1 alpha-2 country code or 'from_token'
        """
        return self._get("albums", ids=",".join(album_ids), market=market)



from httpx import codes

from tekore._sender import Request, Response
from tekore._sender.error import get_error
from tekore.model import PlayerErrorReason

error_format = """Error in {url}:
{code}: {msg}
"""


def parse_error_reason(response: Response) -> str:
    """Extract error reason from response content."""
    reason = getattr(response, "reason", "")

    if response.content is None:
        return reason

    error = response.content["error"]
    message = error.get("message", reason)
    if "reason" in error:
        message += "\n" + PlayerErrorReason[error["reason"]].value
    return message


def handle_errors(request: Request, response: Response) -> None:
    """Examine response and raise errors accordingly."""
    if codes.is_error(response.status_code):
        error_str = error_format.format(
            url=response.url,
            code=response.status_code,
            msg=parse_error_reason(response),
        )
        error_cls = get_error(response.status_code)
        raise error_cls(error_str, request=request, response=response)



from .base import Identifiable


class AudioFeatures(Identifiable):
    """Features of a track."""

    acousticness: float
    analysis_url: str
    danceability: float
    duration_ms: int
    energy: float
    instrumentalness: float
    key: int
    liveness: float
    loudness: float
    mode: int
    speechiness: float
    tempo: float
    time_signature: int
    track_href: str
    type: str
    uri: str
    valence: float



from __future__ import annotations

from .base import Identifiable
from .serialise import Model, StrEnum
from .track import FullTrack


class RecommendationAttribute(StrEnum):
    """Attributes available in recommendations."""

    acousticness = "acousticness"
    danceability = "danceability"
    duration_ms = "duration_ms"
    energy = "energy"
    instrumentalness = "instrumentalness"
    key = "key"
    liveness = "liveness"
    loudness = "loudness"
    mode = "mode"
    popularity = "popularity"
    speechiness = "speechiness"
    tempo = "tempo"
    time_signature = "time_signature"
    valence = "valence"


class RecommendationSeed(Identifiable):
    """Recommendation seeds."""

    afterFilteringSize: int  # noqa: N815
    afterRelinkingSize: int  # noqa: N815
    href: str | None
    initialPoolSize: int  # noqa: N815
    type: str


class Recommendations(Model):
    """Track recommendations."""

    seeds: list[RecommendationSeed]
    tracks: list[FullTrack]



from __future__ import annotations

from tekore._auth import scope
from tekore._client.base import SpotifyBase
from tekore._client.chunked import chunked, join_lists, return_none
from tekore._client.decor import maximise_limit, scopes, send_and_process
from tekore._client.process import nothing, single
from tekore.model import FullArtistCursorPaging


class SpotifyFollow(SpotifyBase):
    """Follow API endpoints."""

    @scopes(optional=[scope.playlist_read_private])
    @chunked("user_ids", 2, 5, join_lists)
    @send_and_process(nothing)
    def playlist_is_following(
        self, playlist_id: str, user_ids: list[str]
    ) -> list[bool]:
        """
        Check if users are following a playlist.

        Parameters
        ----------
        playlist_id
            playlist ID
        user_ids
            list of user IDs, max 5 without chunking

        Returns
        -------
        list[bool]
            follow statuses in the same order that the user IDs were given
        """
        return self._get(
            f"playlists/{playlist_id}/followers/contains", ids=",".join(user_ids)
        )

    @scopes([scope.playlist_modify_public], [scope.playlist_modify_private])
    @send_and_process(nothing)
    def playlist_follow(self, playlist_id: str, public: bool = True) -> None:
        """
        Follow a playlist as current user.

        Parameters
        ----------
        playlist_id
            playlist ID
        public
            follow publicly
        """
        payload = {"public": public}
        return self._put(f"playlists/{playlist_id}/followers", payload=payload)

    @scopes([scope.playlist_modify_public], [scope.playlist_modify_private])
    @send_and_process(nothing)
    def playlist_unfollow(self, playlist_id: str) -> None:
        """
        Unfollow a playlist as current user.

        Parameters
        ----------
        playlist_id
            playlist ID
        """
        return self._delete(f"playlists/{playlist_id}/followers")

    @scopes([scope.user_follow_read])
    @send_and_process(single(FullArtistCursorPaging, from_item="artists"))
    @maximise_limit(50)
    def followed_artists(
        self, limit: int = 20, after: str | None = None
    ) -> FullArtistCursorPaging:
        """
        Get artists followed by the current user.

        Parameters
        ----------
        limit
            the number of items to return (1..50)
        after
            the last artist ID retrieved from the previous request
        """
        return self._get("me/following", type="artist", limit=limit, after=after)

    @scopes([scope.user_follow_read])
    @chunked("artist_ids", 1, 50, join_lists)
    @send_and_process(nothing)
    def artists_is_following(self, artist_ids: list[str]) -> list[bool]:
        """
        Check if current user follows artists.

        Parameters
        ----------
        artist_ids
            list of artist IDs, max 50 without chunking

        Returns
        -------
        list[bool]
            follow statuses in the same order that the artist IDs were given
        """
        return self._get(
            "me/following/contains", type="artist", ids=",".join(artist_ids)
        )

    @scopes([scope.user_follow_modify])
    @chunked("artist_ids", 1, 50, return_none)
    @send_and_process(nothing)
    def artists_follow(self, artist_ids: list[str]) -> None:
        """
        Follow artists as current user.

        Parameters
        ----------
        artist_ids
            list of artist IDs, max 50 without chunking
        """
        return self._put("me/following", type="artist", ids=",".join(artist_ids))

    @scopes([scope.user_follow_modify])
    @chunked("artist_ids", 1, 50, return_none)
    @send_and_process(nothing)
    def artists_unfollow(self, artist_ids: list[str]) -> None:
        """
        Unfollow artists as current user.

        Parameters
        ----------
        artist_ids
            list of artist IDs, max 50 without chunking
        """
        return self._delete("me/following", type="artist", ids=",".join(artist_ids))

    @scopes([scope.user_follow_read])
    @chunked("user_ids", 1, 50, join_lists)
    @send_and_process(nothing)
    def users_is_following(self, user_ids: list[str]) -> list[bool]:
        """
        Check if current user follows users.

        Parameters
        ----------
        user_ids
            list of user IDs, max 50 without chunking

        Returns
        -------
        list[bool]
            follow statuses in the same order that the user IDs were given
        """
        return self._get("me/following/contains", type="user", ids=",".join(user_ids))

    @scopes([scope.user_follow_modify])
    @chunked("user_ids", 1, 50, return_none)
    @send_and_process(nothing)
    def users_follow(self, user_ids: list[str]) -> None:
        """
        Follow users as current user.

        Parameters
        ----------
        user_ids
            list of user IDs, max 50 without chunking
        """
        return self._put("me/following", type="user", ids=",".join(user_ids))

    @scopes([scope.user_follow_modify])
    @chunked("user_ids", 1, 50, return_none)
    @send_and_process(nothing)
    def users_unfollow(self, user_ids: list[str]) -> None:
        """
        Unfollow users as current user.

        Parameters
        ----------
        user_ids
            list of user IDs, max 50 without chunking
        """
        return self._delete("me/following", type="user", ids=",".join(user_ids))



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Coroutine
from dataclasses import dataclass


@dataclass
class Request:
    """Wrapper for parameters of a HTTP request."""

    method: str
    url: str
    params: dict | None = None
    headers: dict | None = None
    data: dict | None = None
    json: dict | None = None
    content: str | None = None


@dataclass
class Response:
    """Wrapper for result of a HTTP request."""

    url: str
    headers: dict
    status_code: int
    content: dict | None


class Sender(ABC):
    """Sender interface for requests."""

    def __repr__(self) -> str:
        return type(self).__name__ + "()"

    @abstractmethod
    def send(self, request: Request) -> Response | Coroutine[None, None, Response]:
        """
        Send a request.

        Parameters
        ----------
        request
            request to send

        Returns
        -------
        Response
            resulting response
        """

    @property
    @abstractmethod
    def is_async(self) -> bool:
        """Sender asynchronicity mode."""

    @abstractmethod
    def close(self) -> None | Coroutine[None, None, None]:
        """Close underlying client."""



from __future__ import annotations

from collections.abc import Sequence

from .serialise import Model


class Paging(Model):
    """Paging base."""

    href: str
    items: Sequence[Model | None]
    limit: int
    next: str | None


class OffsetPaging(Paging):
    """
    Offset paging base.

    Paging that can be navigated both forward and back.
    """

    total: int
    offset: int
    previous: str | None


class Cursor(Model):
    """Data cursor."""

    after: str | None


class CursorPaging(Paging):
    """
    Cursor paging base.

    Paging that can be navigated only forward following the cursor.
    """

    cursors: Cursor



from __future__ import annotations

from collections.abc import Callable

from tekore.model import Model


def nothing(json: dict) -> dict:
    """Pass value without doing anything."""
    return json


def top_item(item: str) -> Callable:
    """Return ``item`` from top level of dict."""

    def post_func(json: dict) -> dict:
        return json[item]

    return post_func


def single(type_: type[Model], from_item: str | None = None) -> Callable:
    """
    Unpack dict or items in ``from_item`` into single constructor.

    If dict or ``from_item`` is None - does nothing and returns None.
    """

    def post_func(json: dict) -> Model | None:
        json = json if from_item is None else json[from_item]
        return type_(**json) if json is not None else None

    return post_func


def model_list(type_: type[Model], from_item: str | None = None) -> Callable:
    """Unpack items inside ``from_item`` of dict into constructors."""

    def post_func(json: dict) -> list[Model | None]:
        json = json if from_item is None else json[from_item]
        return [type_(**i) if i is not None else None for i in json]

    return post_func


def multiple(*args: Callable) -> Callable:
    """Run json dict through multiple processors."""

    def post_func(json: dict) -> tuple:
        return tuple(processor(json) for processor in args)

    return post_func
